"""DnaWeaver Reports implements reporting methods for Dna Weaver outputs."""

from .plotting import (plot_supply_graph, matplotlib_figure_to_svg_base64_data,
                       plot_assembly_graph, plot_assembly_blocks,
                       plot_assembly_timeline,
                       plot_decomposition_graph,
                       matplotlib_figure_to_file_string,
                       autocolor_quote_sources,
                       plot_cost_profiles)
from .reports import (make_pdf_report, make_spreadsheet_sequences_report,
                      make_html_report, make_folder_report)
from .JsonQuote import JsonQuote


from .biotools import gc_content, reverse_complement
from Bio import Restriction
import re

class NoPatternConstraint:
    """Class of callables (sequence)-> True/False whether the sequence contains
    the pattern.

    Can be useful for defining constraints in DNA assembly methods or
    DNA providers.

    The interest of having this as a class is that a DnaSource using this
    constraint can be displayed as a string with the pattern appearing
    explicitly, which would not be the case for a function
    """

    def __init__(self, pattern=None, enzyme=None, is_regex=False,
                 with_revcomp=True):

        self.biopython_enzyme = None
        if enzyme is not None:
            if enzyme in Restriction.__dict__:
                biopython_enzyme = Restriction.__dict__[enzyme]
                if all([c in 'ATGC' for c in biopython_enzyme.site]):
                    pattern = biopython_enzyme.site
                else:
                    self.biopython_enzyme = biopython_enzyme
            else:
                raise ValueError("Unknown enzyme: %s" % enzyme)
        self.enzyme = enzyme
        self.pattern = pattern
        self.is_regex = is_regex
        self.with_revcomp = with_revcomp
        if self.with_revcomp and self.pattern:
            self.rev_pattern = reverse_complement(pattern)

    def __call__(self, sequence):

        if self.biopython_enzyme is not None:
            return self.biopython_enzyme.search(sequence) == []

        if self.is_regex:
            cm_pattern = re.compile(self.pattern)
            if (cm_pattern.search(sequence) is not None):
                if self.with_revcomp:
                    sequence_rev = reverse_complement(sequence)
                    return (cm_pattern.search(sequence_rev) is not None)
                else:
                    return True
            else:
                return False
        else:
            if self.pattern not in sequence:
                if self.with_revcomp:
                    return (self.rev_pattern not in sequence)
                else:
                    return True
            else:
                return False

    def __repr__(self):
        return "No pattern '%s'" % (self.pattern)

class SequenceLengthConstraint:

    def __init__(self, min_length=0, max_length=None):
        self.min_length = min_length
        self.max_length = max_length

    def __call__(self, sequence):
        L = len(sequence)
        upper_bound = self.max_length if self.max_length is not None else L
        return self.min_length <= L <= upper_bound

    def __str__(self):
        left_side = ("" if (self.min_length == 0) else
                     ("%d < " % self.min_length))
        right_side =("" if (self.max_length is None) else
                     (" < %d" % self.max_length))
        return left_side + "length" + right_side


class GcContentConstraint:

    def __init__(self, min_gc=0, max_gc=1.0):
        self.min_gc = min_gc
        self.max_gc = max_gc

    def __call__(self, sequence):
        return self.min_gc <= gc_content(sequence) <= self.max_gc

    def __str__(self):
        left_side = ("" if (self.min_gc == 0) else
                     ("%.01f" % (self.min_gc*100) + "% < "))
        right_side =("" if (self.max_gc == 1) else
                     (" < %.01f" % (self.max_gc*100) + "%"))
        return left_side + "GC" + right_side

class PerBasepairPricing:

    def __init__(self, per_basepair_price):
        self.per_basepair_price = per_basepair_price
        self.min_basepair_price = per_basepair_price

    def __call__(self, sequence):
        return len(sequence) * self.per_basepair_price

    def __str__(self):
        return "$%.03f/bp" % self.per_basepair_price

class FixedPricing:

    def __init__(self, fixed_price):
        self.fixed_price = fixed_price
        self.min_basepair_price = 0

    def __call__(self, sequence):
        return self.fixed_price

    def __str__(self):
        return "$%.03f/order" % self.fixed_price


"""DnaWeaver, a Python package for optimal DNA sequence decomposition and
ordering."""

from .dna_sources import (CommercialDnaOffer,
                          DnaSourcesComparator,
                          DnaAssemblyStation,
                          PcrOutStation,
                          PartsLibrary,
                          GoldenGatePartsLibrary)
from .DnaAssemblyMethod import (GibsonAssemblyMethod,
                                BuildAGenomeAssemblyMethod,
                                GoldenGateAssemblyMethod)
from .OverhangSelector import (OverhangSelector,
                               FixedSizeOverhangSelector,
                               TmOverhangSelector)
from .biotools import (random_dna_sequence, reverse_complement,
                       string_to_sequence, load_record, gc_content)
from .constraints import (NoPatternConstraint, PerBasepairPricing,
                          SequenceLengthConstraint, GcContentConstraint,
                          FixedPricing)
from .supply_network_from_json import supply_network_from_json


import subprocess
import tempfile
import time
import os
import re
try:
    from StringIO import StringIO
except ImportError:  # python 3
    from io import StringIO

from Bio.Blast import NCBIXML
from Bio.Seq import Seq
from Bio import Restriction
from Bio import SeqIO
import numpy as np



def complement(dna_sequence):
    """Return the complement of the DNA sequence.

    For instance ``complement("ATGCCG")`` returns ``"TACGGC"``.
    Uses BioPython for speed.
    """
    return str(Seq(dna_sequence).complement())


def reverse_complement(sequence):
    """Return the reverse-complement of the DNA sequence.

    For instance ``complement("ATGCCG")`` returns ``"GCCGTA"``.
    Uses BioPython for speed.
    """
    return complement(sequence)[::-1]


def random_dna_sequence(length, probas=None, seed=None):
    """Return a random DNA sequence ("ATGGCGT...") with the specified length.

    Parameters
    ----------

    length
      Length of the DNA sequence.

    proba
      Frequencies for the different nucleotides, for instance
      ``probas={"A":0.2, "T":0.3, "G":0.3, "C":0.2}``.
      If not specified, all nucleotides are equiprobable (p=0.25).

    seed
      The seed to feed to the random number generator. When a seed is provided
      the random results depend deterministically on the seed, thus enabling
      reproducibility

    """
    if seed is not None:
        np.random.seed(seed)
    if probas is None:
        sequence = np.random.choice(list("ATCG"), length)
    else:
        bases, probas = zip(*probas.items())
        sequence = np.random.choice(bases, length, p=probas)
    return "".join(sequence)


def load_record(filename, name="unnamed"):
    """Load a sequence file (genbank or fasta) as  a Biopython record."""
    if filename.lower().endswith(("gb", "gbk")):
        record = SeqIO.read(filename, "genbank")
    elif filename.lower().endswith(('fa', 'fasta')):
        record = SeqIO.read(filename, "fasta")
    else:
        raise ValueError('Unknown format for file: %s' % filename)
    record.id = name
    record.name = name.replace(" ", "_")[:20]
    return record

def blast_sequence(sequence, blast_db=None, subject=None, word_size=4,
                   perc_identity=80, num_alignments=1000, num_threads=3,
                   use_megablast=True, ungapped=True):
    """Return a Biopython BLAST record of the given sequence BLASTed
    against the provided database.

    Parameters
    ----------

    sequence
      An ATGC sequence

    blast_db
      Path to a BLAST database

    subject
      Either a path to a fasta (.fa) file or an ATGC string. Subject to blast
      against.

    word_size
      Word size to use in the blast

    perc_identity
      Minimal percentage of identical nucleotides in a match for it to be kept

    num_alignments
      Number of alignments to keep

    num_threads
      Number of threads for the BLAST

    use_megablast
      Whether to use Megablast.

    ungapped
      No-gaps matches only ?

    Examples
    --------

    >>> blast_record = blast_sequence("ATTGTGCGTGTGTGCGT", "blastdb/ecoli")
    >>> for alignment in blast_record.alignments:
    >>>     for hit in alignment.hsps:
    >>>         print (hit.identities)
    """

    xml_file, xml_name = tempfile.mkstemp(".xml")
    fasta_file, fasta_name = tempfile.mkstemp(".fa")
    with open(fasta_name, "w+") as f:
        f.write(">seq\n" + sequence)

    if subject is not None:
        close_subject = True
        if not subject.endswith(".fa"):
            remove_subject = True
            subject_file, fasta_subject_name = tempfile.mkstemp(".fa")
            with open(fasta_subject_name, "w+") as f:
                f.write(">subject\n" + subject)
            subject = fasta_subject_name
        else:
            remove_subject = False
    else:
        close_subject = False
    p = subprocess.Popen([
        "blastn", "-out", xml_name,
        "-outfmt", "5",
        "-num_alignments", str(num_alignments),
        "-query", fasta_name] +
        (["-db", blast_db] if blast_db is not None
         else ['-subject', subject]) +
        (["-ungapped"] if ungapped else []) +
        (["-task", "megablast"] if use_megablast else []) + [
        "-word_size", str(word_size),
        "-num_threads", str(num_threads),
        "-dust", "no",
        "-evalue", "0.01",
        "-perc_identity", str(perc_identity)
    ], close_fds=True, stderr=subprocess.PIPE)
    res, blast_err = p.communicate()
    p.wait()
    error = None
    for i in range(3):
        try:
            with open(xml_name, "r") as f:
                res = list(NCBIXML.parse(f))
                os.fdopen(xml_file, 'w').close()
                os.fdopen(fasta_file, 'w').close()
                os.remove(xml_name)
                os.remove(fasta_name)

                if close_subject:
                    open(subject, 'w').close()
                    if remove_subject:
                        os.remove(subject)
                if len(res) == 1:
                    return res[0]
                else:
                    return res
            break
        except ValueError as err:
            error = err
            time.sleep(0.1)
    else:
        raise ValueError("Problem reading the blast record: " + str(error))


def perfect_match_locations_in_hsp(hsp, span_cutoff=10):
    """Return the locations of perfect matches in a BLAST HSP.
    
    Only locations with a span above span_cutoff are kept.
    """
    if hsp.align_length < span_cutoff:
        return []
    arr = np.frombuffer(hsp.match.encode(), dtype='uint8')
    indices = [0] + list((arr != 124).nonzero()[0]) + [len(arr)]
    return [
        (start + hsp.query_start, end + hsp.query_start)
        for start, end in zip(indices, indices[1:])
        if end - start >= span_cutoff
    ]

def largest_common_substring(query, target, max_overhang):
    """Return the largest common substring between `query` and `target`.

    If the common substring is too much smaller than `query` False is returned,
    else the location `(start, end)` of the substring in `target` is returned.

    Parameters:
    -----------

    query (str)
      The sequence to be found in target (minus some overhangs possibly)

    target (str)
      The sequence in which to find `query`

    max_overhang
      Maximal size allowed for the flanking regions of `query` that would
      not be contained in `target`.

    Notes:
    ------

    This is intended for finding whether `query` can be extracted from `target`
    using PCR. See the PcrOutStation implementation in DNASource.py.

    """
    start, end = max_overhang, len(query) - max_overhang
    if query[start:end] not in target:
        return False
    while (start >= 0) and (query[start:end] in target):
        start -= 1
    start += 1
    while (end < len(query)) and (query[start:end] in target):
        end += 1
    end -= 1
    return start, end


def gc_content(sequence, window_size=None):
    """Compute global or local GC content.

    Parameters
    ----------

    sequence
      An ATGC DNA sequence (upper case!)

    window_size
      If provided, the local GC content for the different sliding windows of
      this size is returned, else the global GC content is returned.

    Returns
    --------

      A number between 0 and 1 indication the proportion
      of GC content. If window_size is provided, returns
      a list of len(sequence)-window_size values indicating
      the local GC contents (sliding-window method). The i-th value
      indicates the GC content in the window [i, i+window_size]
    """
    # The code is a little cryptic but speed gain is 300x
    # compared with pure-python string operations

    arr = np.fromstring(sequence + "", dtype="uint8")
    arr_GCs = (arr == 71) | (arr == 67)  # 67=C, 71=G

    if window_size is None:
        return 1.0 * arr_GCs.sum() / len(sequence)
    else:
        cs = np.cumsum(arr_GCs)
        a = cs[window_size - 1:]
        b = np.hstack([[0], cs[:-window_size]])
        return 1.0 * (a - b) / window_size

def find_enzyme_sites(sequence, enzyme_name, padding=0, padding_nuc="A"):
    padding = padding * padding_nuc
    sequence = Seq(padding + sequence + padding)
    return Restriction.__dict__[enzyme_name].search(sequence)

def string_to_sequence(string):
    """Convert a string of a fasta, genbank... into a simple ATGC string.

    Can also be used to detect a format.

    """
    matches = re.match("([ATGC][ATGC]*)", string)
    if (matches is not None) and (matches.groups()[0] == string):
        return (string, "ATGC")

    for fmt in ("fasta", "genbank"):
        try:
            stringio = StringIO(string)
            return (str(SeqIO.read(stringio, fmt).seq), fmt)
        except:
            pass
    raise ValueError("Invalid sequence format")

def file_to_sequence(filename):
    """Import a file in fasta, genbank... as a simple ATGC string."""
    with open(filename, "r") as f:
        return string_to_sequence(f.read())

def gc_content_to_tm(seq_length, gc_content, gc_tm=4, nongc_tm=2):
    return seq_length * (gc_content * gc_tm + (1 - gc_content) * nongc_tm)

def make_blast_db(fasta_input, target):
  proc = subprocess.Popen([
      "makeblastdb",
      "-in", fasta_input,
      "-dbtype", "nucl",
      "-out", target
  ])
  proc.wait()


import networkx as nx
from .dna_sources import (CommercialDnaOffer, DnaAssemblyStation, PartsLibrary,
                          PcrOutStation, DnaSourcesComparator,
                          GoldenGatePartsLibrary)
DEFAULT_DNA_SOURCES_DICT = {
    'commercial':  CommercialDnaOffer,
    'assembly': DnaAssemblyStation,
    'library': PartsLibrary,
    'golden_gate_library': GoldenGatePartsLibrary,
    'pcr': PcrOutStation,
    'comparator': DnaSourcesComparator,
    'main': DnaSourcesComparator
}

def _sort_suppliers(graph_data):
    supply_graph = nx.DiGraph([
        (supplier, supplier_id)
        for supplier_id, supplier_data in graph_data.items()
        for supplier in supplier_data['suppliers']
    ])
    sorted_suppliers = []
    level = 0
    levels = {}
    while len(supply_graph):
        level += 1
        independant_suppliers = [
            n for n in supply_graph.nodes()
            if len(nx.ancestors(supply_graph, n)) == 0
        ]
        for supp in independant_suppliers:
            levels[supp] = level
        sorted_suppliers.extend(independant_suppliers)
        supply_graph.remove_nodes_from(independant_suppliers)
    return sorted_suppliers, levels


def supply_network_from_json(graph_data, dna_sources_dict='default'):
    """

    Returns
    -------

    levels, suppliers_dict, main_id
    """
    if dna_sources_dict == 'default':
        dna_sources_dict = DEFAULT_DNA_SOURCES_DICT

    sorted_suppliers, levels = _sort_suppliers(graph_data)
    main_id = sorted_suppliers[-1]
    suppliers_dict = {}
    for supplier_id in sorted_suppliers:
        supplier_data = graph_data[supplier_id]
        supplier_data['parameters']["name"] = supplier_data["name"]
        supplier_data['parameters']['suppliers'] = [
            suppliers_dict[supp_id]
            for supp_id in supplier_data['suppliers']
        ]
        supplier_class = dna_sources_dict[supplier_data["type"]]
        supplier = supplier_class.from_dict(supplier_data['parameters'])
        supplier.id = supplier_id
        suppliers_dict[supplier_id] = supplier
    return levels, suppliers_dict, main_id


from copy import deepcopy
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Alphabet import DNAAlphabet
from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation

try:  # Python2 vs. Python3 ways of dealing with file-like strings
    from StringIO import StringIO
    USE_BYTES = False
except ImportError:
    from io import StringIO, BytesIO
    USE_BYTES = True

class ObjectDict(dict):

    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        dict.__getattribute__(self, key)

    def __setattr__(self, key, value):
        self[key] = value
        self.__dict__[key] = value

    @staticmethod
    def from_dict(d):
        obj = ObjectDict({
            key: ([
                ObjectDict.from_dict(e)
                if isinstance(e, dict) else e
                for e in value
            ]
                if isinstance(value, (list, tuple))
                else (ObjectDict.from_dict(value)
                      if isinstance(value, dict)
                      else value))
            for (key, value) in d.items()
        })
        for key, value in obj.items():
            sanitized_key = key.replace(" ", "_").replace(".", "_")
            obj.__dict__[sanitized_key] = value
        return obj


class JsonQuote:

    def __init__(self, tree, sources):
        self.tree = ObjectDict.from_dict(tree)
        self.sources = ObjectDict.from_dict(sources)

    @staticmethod
    def from_dnaweaver_quote(quote):
        tree = quote.assembly_tree_as_dict()
        sources = quote.source.dict_supply_graph()
        return JsonQuote(tree, sources)

    def to_quotes_list(self):
        tree = deepcopy(self.tree)
        nodes = []

        def rec(node, depth=0):
            if node.get("_visited", False):
                return
            node["_visited"] = True
            assembly_plan = node.get("assembly_plan", [])
            node["children"] = [
                n["id"] for n in assembly_plan
            ]
            nodes.append(node)
            for other in sorted(assembly_plan,
                                key=lambda n: n["segment_start"]):
                rec(other)
        rec(tree)
        return nodes

    def quote_tree_to_SeqRecord(self, record=None, record_id=None):
        """Return a Biopython seqrecord of the quote.

        >>> record = to_SeqRecord(solution)
        >>> # Let's plot with DnaVu:
        >>> from dnavu import create_record_plot
        >>> from bokeh.io import output_file, show
        >>> output_file("view.html")
        >>> plot = create_record_plot(record)
        >>> show(plot)
        """
        if record_id is None:
            record_id = self.id
        if record is None:
            record = SeqRecord(Seq(self.sequence, DNAAlphabet()), id=record_id)
        else:
            record = deepcopy(record)

        if self.assembly_plan is not None:
            features = [
                SeqFeature(
                    FeatureLocation(segment[0], segment[1], 1),
                    type="Feature",
                    qualifiers={
                        "label": "%s - From %s" % (quote.id, quote.source),
                        "name": quote.id,
                        "source": quote.source,
                        "price": quote.price,
                        "lead_time": quote.lead_time
                    }
                )
                for segment, quote in self.assembly_plan.items()
            ]
            record.features = features + record.features
        return record

    def quote_tree_to_genbank(self, filename=None, filehandle=None,
                              record=None, record_id=None):
        record = self.to_SeqRecord(record=record, record_id=record_id)
        if filename is not None:
            with open(filename, "w+") as f:
                SeqIO.write(record, f, "genbank")
        else:
            output = StringIO()
            SeqIO.write(record, output, "genbank")
            return output.getvalue()



"""Global settings for DnaWeaver

Examples
--------

>>> import dnaweaver.config as config
>>> fontawesome_path = config["fontawesome-ttf-path"]
>>> # To change that path globally for DnaWeaver:
>>> config["fontawesome-ttf-path"] = "some/new/path.ttf"

"""

import os

this_path = os.path.join(os.path.dirname(os.path.realpath(__file__)))
fonts_path = os.path.join(this_path, "fonts")
fontawesome_path = os.path.join(fonts_path,  "fontawesome-webfont.ttf")
opensans_path = os.path.join(fonts_path, 'OpenSans-Light.ttf')
SETTINGS = {
    "fontawesome-ttf-path": fontawesome_path,
    "OpenSans-ttf-path": opensans_path,
    "template_path": os.path.join(this_path, "templates")
}



""" This module contains overhangs selectors for methods such as
Gibson Assembly, Golden Gate assembly, recombination in yeast."""

import numpy as np
from functools import lru_cache
try:
    import primer3
    PRIMER3_AVAILABLE = True
except:
    PRIMER3_AVAILABLE = False



class OverhangSelector:
    """Base class for overhang selectors such as TmOverhangSelector.

    These selectors return an ideal overhang subsegment around a given sequence
    location. They can also be used to filter out locations in the sequence as
    non-potential cutting sites for sequence decomposition.
    """

    def location_filter_method(self, sequence):
        """Return a filter function f(location) => True/False.

        The result is True iff the location is a valid cutting site"""
        if self.has_location_filter:
            def f(index):
                return self.filter_location(sequence, index)
            return f
        else:
            return None

    @property
    def has_location_filter(self):
        return hasattr(self, 'filter_location')

    def compute_sequence_fragment(self, sequence, segment):
        """Compute the fragment equal to the sequence segment + overhangs."""
        start, end = segment
        if start == 0:
            fragment_start = 0
        else:
            fragment_start = self.compute_overhang_location(sequence, start)[0]
        if end == len(sequence):
            fragment_end = len(sequence)
        else:
            fragment_end = self.compute_overhang_location(sequence, end)[1]
        fragment = sequence[fragment_start: fragment_end]
        return self.left_addition + fragment + self.right_addition


    def compute_overhang_sequence(self, sequence, index):
        """Return the sequence of the selected overhang at the given index."""
        start, end = self.compute_overhang_location(sequence, index)
        return sequence[start: end]

    @staticmethod
    def get_segment_coordinates(center, segment_length, sequence_length):
        """Return max(0, c - s/2) - min(L, c + L/2).

        Where c=center, s=segment_length, L=sequence_length.
        """
        half = int(segment_length / 2)
        start = max(0, min(center - half, sequence_length - segment_length))
        end = start + segment_length
        return start, end


class FixedSizeOverhangSelector(OverhangSelector):
    """Selects overhangs of a constant size.

    Great for methods involving large homology regions where melting
    temperature matters less.
    """

    def __init__(self, overhang_size=100, left_addition='', right_addition=''):
        self.overhang_size = overhang_size
        self.left_addition = left_addition
        self.right_addition = right_addition


    def compute_overhang_location(self, sequence, index):
        return self.get_segment_coordinates(index, self.overhang_size,
                                            len(sequence))

class TmOverhangSelector(OverhangSelector):
    """Selects overhangs with melting temperature constraints.

    Given a location in a sequence, this selector will return an overhang
    (subsegment of the sequence) around this position, with size and
    temperature constraints.

    Can be used for Gibson Assembly, Golden Gate Assembly, etc.

    This class also implements methods to quickly compute all overhangs in a
    sequence. The result is then cached for speed.

    Parameters
    ----------
    min_size
      Minimal length of the overhang, in nucleotides

    max_size
      Maximal length of the overhang, in nucleotides

    min_tm
      Minimal melting temp allowed for overhangs

    max_tm
      Maximal melting temp allowed for overhangs

    precompute_overhangs
      If True, the whole sequence will be analyzed to find the best overhang
      for each cutting point (and find cutting points which do not allow
      overhangs), using the AT/GC=2/4C heuristic for melting temperatures.

    primer3_params
      If ``precompute_overhangs`` is False and ``Primer3`` is installed, the
      melting temperature is computed
    """

    def __init__(self, min_size=18, max_size=22, min_tm=55, max_tm=65,
                 precompute_overhangs=True, left_addition='',
                 right_addition='', **primer3_params):
        """Initialize."""
        self.min_size = min_size
        self.max_size = max_size
        self.precompute_overhangs = precompute_overhangs
        self.primer3_params = primer3_params
        self.min_tm = min_tm
        self.max_tm = max_tm
        self.middle_tm = 0.5 * (self.min_tm + self.max_tm)
        self.middle_size = int(0.5 * (self.min_size + self.max_size))
        self.left_addition = left_addition
        self.right_addition = right_addition

    def filter_location(self, sequence, index):
        """Return whether the sequence has a valid overhang at this index."""
        if index == len(sequence):
            index = index - 1
        if self.precompute_overhangs:
            return (self.compute_all_overhangs(sequence)[index] is not None)
        else:
            return self.compute_optimal_overhang(sequence, index)[1]

    def compute_overhang_location(self, sequence, index):
        """Return the location (start, stop) of the selected overhang."""
        if index == len(sequence):
            index = index - 1
        if self.precompute_overhangs:
            # calling a result that should be computed once then cached
            return self.compute_all_overhangs(sequence)[index]
        else:
            return self.compute_optimal_overhang(sequence, index)[0]

    def compute_optimal_overhang(self, sequence, index, init_size=None):
        """Return the best overhang around some location.

        This function will be shortcut if ``precompute_overhangs`` is True,
        as it is slow to compute the best overhang for each location
        separately.
        """

        if init_size is None:
            init_size = self.middle_size

        def f(ovh_size):
            start, end = self.get_segment_coordinates(index, ovh_size,
                                                      len(sequence))
            tm = self.compute_tm(sequence[start: end])
            return (start, end), tm
        best_overhang, best_tm = f(init_size)
        if best_tm < self.middle_tm:
            ovh_sizes = range(init_size, self.max_size)
        else:
            ovh_sizes = range(init_size, self.min_size, -1)
        for ovh_size in ovh_sizes:
            new_overhang, new_tm = f(ovh_size)
            if abs(new_tm - self.middle_tm) < abs(best_tm - self.middle_tm):
                best_overhang, best_tm = new_overhang, new_tm
            else:
                break
        return best_overhang, (self.min_tm < best_tm < self.max_tm)

    @lru_cache(maxsize=3)
    def compute_all_overhangs(self, sequence):
        """Quickly compute all overhangs in the sequence.

        This function uses the heuristic {A, T}=2degC, {G, C}=4degC to compute
        melting temperatures.

        This function uses vectorial operations for speed. The results are also
        cached.
        """

        if self.primer3_params == {}:
            lmin, lmax = self.min_size, self.max_size
            table = np.zeros((lmax + 1 - lmin, len(sequence)))

            # super fast GC computing using the trick 67=C, 71=G in unicode.
            seq_array = np.fromstring(sequence + "", dtype="uint8")
            cumsum = np.cumsum(2 + 2 * ((seq_array == 71) | (seq_array == 67)))
            # cumsum = np.cumsum([
            #     4 if nuc in "GC" else 2
            #     for nuc in sequence
            # ])
            for i, oh_size in enumerate(range(lmin, lmax + 1)):
                arr = cumsum[oh_size:] - cumsum[:-oh_size]
                start = int(oh_size / 2)
                end = start + len(arr)
                table[i, start:end] = arr
                table[i, :start] = table[i, start]
                table[i, end:] = table[i, end-1]
            scores = - (table - self.min_tm) * (table - self.max_tm)
            best_sizes_indices = scores.argmax(axis=0)
            best_sizes = lmin + best_sizes_indices
            validities = np.choose(best_sizes_indices, scores) >= 0
            osizes_and_validities = zip(best_sizes, validities)
        else:
            osizes_and_validity = [self.compute_optimal_overhang(sequence, 0)]
            for i in range(1, len(sequence)):
                init_size = osizes_and_validity[-1][0]
                ovh, val = self.find_optimal_overhang(sequence, i, init_size)
                osizes_and_validity.append((len(ovh), val))

        return [
          None if not valid
          else self.get_segment_coordinates(i, ovh_size, len(sequence))
          for i, (ovh_size, valid) in enumerate(osizes_and_validities)
        ]


    def compute_tm(self, sequence):
        """Return the melting temp of the sequence.

        If Primer3 is available, it's internal melting temperature calculator
        is used with ``self.primer3_params`` used as parameters.
        Else the heuristic AT/GC=2/4C is used.
        """
        if self.params == {}:
            return sum([
                4 if c in "GC" else 2
                for c in sequence
            ])
        if not PRIMER3_AVAILABLE:
            raise ImportError("Melting temperature computation with '%s' "
                              "Requires Primer3 installed." %
                              self.primer3_params.get('method',
                                                      "[unknown method]"))
        return primer3.calcTm(sequence, **self.primer3_params)



from Bio import SeqIO
from .DnaSource import DnaSource
from ..DnaQuote import DnaQuote

class PartsLibrary(DnaSource):
    """Class for collections of ready-to-assemble parts.


    This class is admitedly under-developed and could be expanded-subclassed
    to accomodate with the different kinds of registries etc.
    """
    class_description = "Parts Library"
    operation_type = "library"
    report_fa_symbol = u""
    report_fa_symbol_plain = "book"

    report_color = "#feeefe"
    collections_by_id = {}
    library_classes = {}

    def __init__(self, name, parts_dict=None, fasta_file=None, memoize=False,
                 price_per_part=0, lead_time=0, sequence_constraints=()):
        self.name = name
        self.price_per_part = price_per_part
        self.lead_time = lead_time
        self.sequence_constraints = sequence_constraints
        if fasta_file is not None:
            parts_dict = {
                record.id: str(record.seq).upper()
                for record in SeqIO.parse(fasta_file, "fasta")
            }
        self.parts_dict = parts_dict
        self.inverted_parts_dict = {v: k for k, v in parts_dict.items()}
        self.sequences_set = set(self.inverted_parts_dict)
        self.memoize = memoize
        self.memoize_dict = {}

    def get_best_price(self, sequence, max_lead_time=None,
                       with_assembly_plan=False):
        """Returns a price-optimal DnaQuote for the given sequence.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quotes

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
       """
        sequence = self.preprocess_sequence(sequence) 
        if sequence in self.sequences_set:
            return DnaQuote(
                self, sequence, accepted=True, price=self.price_per_part,
                lead_time=self.lead_time,
                message="Part: " + self.inverted_parts_dict[sequence]
            )

        return DnaQuote(self, sequence, accepted=False,
                        message="Sequence not in the library")
    def preprocess_sequence(self, sequence):
        """Can be used by subclasses e.g. to anonymize wildcard nucleotides"""
        return sequence

    def additional_dict_description(self):
        return {
            "flanks length": self.flanks_length
        }
    @classmethod
    def from_dict(cls, data):
        parameters = cls.collections_by_id[data['collection']]
        library_class = parameters.pop('library_class')
        if library_class in cls.library_classes:
            library_class = cls.library_classes[library_class]
        def get(param, default):
            return data.get(param, parameters.get(param, default))
        return library_class(
            name= get('name', 'library'),
            price_per_part=get('price_per_part', 0),
            lead_time=get('lead_time', 0),
            parts_dict=get('parts_dict', None),
            fasta_file=get('fasta_file', None),
            memoize=get('memoize', None)
        )


class GoldenGatePartsLibrary(PartsLibrary):
    """Library of parts for Golden Gate Assembly"""
    class_description = "Golden Gate parts library"

    def __init__(self, name, parts_dict=None, fasta_file=None,
                 price_per_part=0, lead_time=0, flanks_length=7, memoize=False,
                 sequence_constraints=()):
        PartsLibrary.__init__(self,  name, parts_dict=parts_dict,
                              fasta_file=fasta_file, memoize=memoize,
                              sequence_constraints=sequence_constraints)
        self.flanks_length = flanks_length

    def suggest_cuts(self, sequence):
        suggested_cuts = []
        # + 2 is because the cut falls in the middle of the 4bp linker:
        flank = self.flanks_length
        for part, part_sequence in self.parts_dict.items():
            segment = part_sequence[flank:-flank]
            i = sequence.find(segment)
            if i != -1:
                suggested_cuts += [i + 2, i + len(segment) - 2]
        return sorted(list(set(suggested_cuts)))
    
    def suggest_segments(self, sequence):
        suggested_segments = []
        # + 2 is because the cut falls in the middle of the 4bp linker:
        flank = self.flanks_length
        for part, part_sequence in self.parts_dict.items():
            segment = part_sequence[flank:-flank]
            i = sequence.find(segment)
            if i != -1:
                L = len(segment)
                suggested_segments.append(((i + 2, i + L - 2), part))
        return sorted(set(suggested_segments))

    @classmethod
    def preprocess_sequence(cls, sequence):
        """Can be used by subclasses e.g. to anonymize wildcard nucleotides"""
        return sequence[:6] + 'N' + sequence[7: -7] + 'N' + sequence[-6:]

    def additional_dict_description(self):
        return {
            "class": "Golden Gate parts library",
            "operation_type": "library",
            "flanks length": self.flanks_length
        }

PartsLibrary.library_classes.update({
    'library': PartsLibrary,
    'golden_gate': GoldenGatePartsLibrary
})


import numpy as np

from ..DnaQuote import DnaQuote
from .DnaSource import DnaSource
from ..tools import functions_list_to_string

from Bio.Restriction import AllEnzymes

from ..constraints import (GcContentConstraint, NoPatternConstraint,
                           PerBasepairPricing, SequenceLengthConstraint)

class CommercialDnaOffer(DnaSource):
    """External/Commercial source of DNA"""

    class_description = "External DNA offer"
    operation_type = "order"
    report_fa_symbol = u""
    report_fa_symbol_plain = "shopping-cart"
    report_color = "#ffeeee"

    def __init__(self, name, pricing, sequence_constraints=(),
                 lead_time=0, memoize=False):
        self.name = name
        self.sequence_constraints = sequence_constraints
        self.pricing = pricing
        self.lead_time = (lead_time if callable(lead_time)
                          else (lambda *a: lead_time))
        self.memoize = memoize
        self.memoize_dict = {}
        if hasattr(pricing, 'min_basepair_price'):
            self.min_basepair_price = pricing.min_basepair_price

    def __repr__(self):
        return self.name

    def get_best_price(self, sequence, max_lead_time=None,
                       with_assembly_plan=False):
        """Returns a price-optimal DnaQuote for the given sequence.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quots

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
        """

        lead_time = self.lead_time(sequence)
        price = self.pricing(sequence)
        accepted = (max_lead_time == np.inf) or (lead_time <= max_lead_time)
        return DnaQuote(self, sequence, accepted=accepted,
                        lead_time=lead_time, price=price)

    def get_best_lead_time_under_price_limit(self, sequence, max_price):

        lead_time = self.lead_time(sequence)
        price = self.pricing(sequence)
        return DnaQuote(self, sequence, price=price, lead_time=lead_time,
                        accepted=price < max_price)

    def additional_dict_description(self):
        constraints = functions_list_to_string(self.sequence_constraints)
        return {
            "price function": functions_list_to_string([self.pricing]),
            "sequence constraints": constraints
        }

    def from_dict(data):
        constraints = []
        if "gc_range" in data:
            mini, maxi = data['gc_range']
            if (mini, maxi) != (0, 100):
                constraints.append(GcContentConstraint(.01 * mini, .01 * maxi))
        if "size_range" in data:
            mini, maxi = data['size_range']
            constraints.append(SequenceLengthConstraint(mini, maxi))
        if 'forbidden' in data:
            ENZYMES = list(str(e) for e in AllEnzymes)
            for pattern in data["forbidden"].split(", "):
                if pattern == '':
                    continue
                enzyme = None
                if pattern in ENZYMES:
                    enzyme = pattern
                    pattern = None
                constraints.append(
                    NoPatternConstraint(pattern=pattern, enzyme=enzyme))

        return CommercialDnaOffer(
            name=data['name'],
            pricing=PerBasepairPricing(data['cost_per_bp']),
            lead_time=data['lead_time'],
            memoize=data.get('memoize', False),
            sequence_constraints=constraints
        )



from ..DnaQuote import DnaQuote
from .DnaSource import DnaSource
from ..constraints import SequenceLengthConstraint
from ..biotools import (reverse_complement, largest_common_substring,
                        blast_sequence, perfect_match_locations_in_hsp)
from ..tools import functions_list_to_string
from .DnaSourcesComparator import DnaSourcesComparator

class PcrOutStation(DnaSource):
    """Class to represent databases of constructs which can be (in part) reused

    A blast database contains the sequences of all available constructs.
    Given a sequence, the PcrOutStation finds whether it is possible to order
    two primers to extract this sequence from the constructs in the BLAST
    database.

    Parameters
    ----------

    name
      Name of the PCR station (e.g. "Lab constructs PCR station")

    primers_dna_source
      DnaSource providing the primers (will typically be an CommercialDnaOffer)

    blast_database

    sequences
      A dictionary {seq_name: sequence_in_atgc}

    pcr_homology_length

    max_overhang_length

    extra_cost

    extra_time

    max_amplicon_length

    blast_word_size

    memoize

    sequence_constraints

    """
    class_description = "PCR-out station"
    operation_type = "PCR"
    report_fa_symbol = u""
    report_fa_symbol_plain = "exchange"
    report_color = "#eeffee"
    dna_banks = {}

    def __init__(self, name, primers_dna_source, blast_database=None,
                 sequences=None, pcr_homology_length=25,
                 max_overhang_length=40, extra_cost=0, extra_time=0,
                 max_amplicon_length=None, blast_word_size=50, memoize=False,
                 sequence_constraints=()):
        self.name = name
        self.blast_database = blast_database
        self.set_suppliers(primers_dna_source)
        self.pcr_homology_length = pcr_homology_length
        self.max_overhang_length = max_overhang_length
        self.extra_time = extra_time
        self.extra_cost = extra_cost
        self.max_amplicon_length = max_amplicon_length
        self.blast_word_size = blast_word_size
        self.sequence_constraints = list(sequence_constraints)
        if max_amplicon_length is not None:
            c = SequenceLengthConstraint(max_length=max_amplicon_length)
            self.sequence_constraints = [c] + self.sequence_constraints
        self.memoize = memoize
        self.memoize_dict = {}
        self.sequences = sequences

    def get_hits(self, sequence):
        """Return the hits of the given sequence against the blast database
        (in Biopython format)"""
        if self.sequences is not None:
            result = []
            for dna_name, seq in self.sequences.items():
                match_coords = largest_common_substring(sequence, seq,
                                                        self.max_overhang_length)
                if match_coords:
                    result.append((dna_name, match_coords, None))
            return result
        else:
            record = blast_sequence(sequence, self.blast_database,
                                    perc_identity=98,
                                    use_megablast=True,
                                    word_size=self.blast_word_size)

            return [
                (al.hit_id + "_h%03d" % i,
                 (hit.query_start, hit.query_end),
                 hit.sbjct)
                for al in record.alignments
                for i, hit in enumerate(al.hsps)
            ]
    def blast_sequence(self, sequence, cutoff=40):
            record = blast_sequence(sequence, self.blast_database,
                                    perc_identity=98,
                                    use_megablast=True,
                                    word_size=self.blast_word_size)
            return [
                (
                    al.hit_id + "_h%03d_%02d" % (i, j),
                    (start, end),
                    sequence[start: end]
                )
                for al in record.alignments
                for i, hit in enumerate(al.hsps)
                for j, (start, end) in enumerate(
                    perfect_match_locations_in_hsp(hit, span_cutoff=cutoff))
            ]

    def get_best_price(self, sequence, max_lead_time=None,
                       with_assembly_plan=False):
        """Return a price-optimal DnaQuote for the given sequence.

        It will find a possible hit in the blast database, find the primers to
        order for the PCR, compute the overall price and lead time, and return
        a quote.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quote

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
        """

        for subject, (hit_start, hit_end), _ in self.get_hits(sequence):

            largest_overhang = max(hit_start, len(sequence) - hit_end)

            if largest_overhang <= self.max_overhang_length:
                primer_l_end = hit_start + self.pcr_homology_length
                primer_left = sequence[:primer_l_end]
                primer_r_end = hit_end - self.pcr_homology_length
                primer_right = reverse_complement(sequence[primer_r_end:])

                primer_max_lead_time = (None if max_lead_time is None else
                                        max_lead_time - self.extra_time)
                quotes = [
                    self.primers_dna_source.get_quote(
                        primer, max_lead_time=primer_max_lead_time
                    )
                    for primer in [primer_left, primer_right]
                ]
                if not all(quote.accepted for quote in quotes):
                    continue  # primers inorderable

                if max_lead_time is not None:
                    overall_lead_time = (max(quote.lead_time
                                             for quote in quotes) +
                                         self.extra_time)
                else:
                    overall_lead_time = None
                total_price = (sum(quote.price for quote in quotes) +
                               self.extra_cost)

                if with_assembly_plan:
                    assembly_plan = {
                        (0, primer_l_end): quotes[0],
                        (primer_r_end, len(sequence)): quotes[1]
                    }
                else:
                    assembly_plan = None

                return DnaQuote(self, sequence, accepted=True,
                                lead_time=overall_lead_time,
                                price=total_price,
                                assembly_plan=assembly_plan,
                                message='From %s' % subject,
                                metadata={"subject": subject,
                                          "location": (hit_start, hit_end)})
        return DnaQuote(self, sequence, accepted=False,
                        message="No valid match found")

    def suggest_cuts(self, sequence):
        if self.sequences is None:
            return []
        suggested_cuts = []
        for name, subseq in self.sequences.items():
            index = sequence.find(subseq)
            if index >= 0:
                suggested_cuts.extend([index, index + len(subseq)])
        return sorted(list(set(suggested_cuts)))

    def prepare_on_sequence(self, sequence):
        """Pre-compute the BLAST of the current sequence against the database.

        Once a pre-blast has been performed, this PcrOutStation becomes
        specialized on that sequence and its subsequences, do not feed it with
        another different sequence. Do `self.sequences=None` to reinitialize
        and de-specialize this PcrOutStation.

        Examples
        --------

        >>> pcr_station = PcrOutStation("some_blast_database")
        >>> top_station = # some assembly station depending on pcr_station
        >>> pcr_station.prepare_on_sequence(my_sequence)
        >>> top_station.get_quote(my_sequence)
        >>> pcr_station.sequences=None # de-specializes the pcr station.
        """
        hits = self.blast_sequence(sequence)
        self.sequences = None  # destroy current pre-blast (used by get_hits)
        self.sequences = {
            subject: seq
            for subject, (start, end), seq in hits
        }

    def additional_dict_description(self):
        constraints = functions_list_to_string(self.sequence_constraints)
        return {
            "BLAST database": self.blast_database,
            "primers dna source": self.primers_dna_source.name,
            "pcr homology length": self.pcr_homology_length,
            "max overhang length": self.max_overhang_length,
            "extra time": self.extra_time,
            "extra cost": self.extra_cost,
            "max amplicon length": self.max_amplicon_length,
            "BLAST word size": self.blast_word_size,
            "sequence constraints": constraints,
        }

    @classmethod
    def from_dict(cls, data):
        if 'dna_bank' in data:
            blast_database = cls.dna_banks[data['dna_bank']]
        else:
            blast_database = data['blast_database']
        return PcrOutStation(
            name=data['name'],
            primers_dna_source=data['suppliers'],
            pcr_homology_length=data['pcr_homology_length'],
            max_overhang_length=data['max_overhang_length'],
            max_amplicon_length=data['max_amplicon_length'],
            extra_cost=data['cost'],
            blast_database=blast_database
        )

    def set_suppliers(self, suppliers):
        if hasattr(suppliers, '__iter__'):
            if len(suppliers) > 1:
                self.primers_dna_source = DnaSourcesComparator(
                    name=self.name + ' comparator', suppliers=suppliers)
            else:
                self.primers_dna_source = suppliers[0]
        else:
            self.primers_dna_source = suppliers



import functools

def memoize(obj):
    """Memoize the function results"""
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]
    return memoizer

def functions_list_to_string(functions):
    def name_function(fun):
        name = str(fun)
        return "unnamed function" if name.startswith("<function") else name
    return " ; ".join([name_function(fun) for fun in functions])



"""Classes to represent assembly strategies in DnaWeaver."""

import itertools as itt
from copy import deepcopy
from collections import defaultdict
import tempfile
import os
import json
try:
    from StringIO import StringIO
except ImportError:  # python 3
    from io import StringIO

from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Alphabet import DNAAlphabet
from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation
from .biotools import blast_sequence


class DnaQuote:
    """Class to represent the Quote returned by a DNA source in response to
    a quote request for a given DNA sequence.

    Parameters
    -----------

    source
      The DnaSource which issued the quote



    sequence

    price

    accepted

    lead_time

    assembly_plan
      A dict of the form { (start1, end1): op1, (start2, end2): op2, ...}
      where the `(start, end)` represent sub-segments of the final sequence and
      `op1, op2` represent the AssemblyOperations that create the corresponding
      fragments

     metadata

     message

     id
       A string identifying uniquely the assembly operation


     deadline
       Deadline for the operation (see the `propagate_deadline` method for this
       class) i.e. in how many time units it should be finished. This is for
       drawing assembly timelines.

    """

    def __init__(self, source, sequence, price=None, accepted=True,
                 lead_time=None, assembly_plan=None, metadata={}, message="",
                 id=None, deadline=None):
        self.source = source
        self.accepted = accepted
        self.price = price
        self.lead_time = lead_time
        self.sequence = sequence
        self.message = message
        self.assembly_plan = assembly_plan
        self.metadata = metadata
        self.id = id
        self.deadline = None

    def __repr__(self):
        if self.accepted:
            infos = "price %.02f" % self.price
            if self.lead_time is not None:
                infos = infos + " - lead_time %0.1f" % self.lead_time
        else:
            infos = "refused: %s" % self.message
        return "From %s - %s - %s" % (self.source, infos, self.message)

    def compute_full_assembly_tree(self, id_prefix="S"):
        """ """

        counter = itt.count()

        def rec(quote):
            if not quote.accepted:
                return quote

            if any([hasattr(quote.source, attr) for attr in [
                   "dna_source", "primers_dna_source"]]):
                if quote.assembly_plan is None:
                    quote = quote.source.get_quote(
                        quote.sequence,
                        max_lead_time=quote.lead_time,
                        with_assembly_plan=True
                    )
                segments = {
                    segment: rec(subquote)
                    for segment, subquote in
                    sorted(quote.assembly_plan.items(),
                           key=lambda item: item[0])
                }
                quote.assembly_plan = segments
            if id_prefix:
                quote.id = id_prefix + "%05d" % next(counter)
            return quote

        rec(self)

        if id_prefix:
            self.id = id_prefix + "%05d" % next(counter)

    def compute_fragments_final_locations(self):
        """Compute the final location of the fragments.

        Examples
        --------

        >>> quote.compute_full_assembly_tree()
        >>> quote.compute_fragments_final_locations()
        >>> some_subquote = print quote.assembly_plan.values()[0]
        >>> print (some_subquote.final_location)
        """
        self.compute_full_assembly_tree()
        quotes = self.tree_as_list()
        quotes_dict = {quote.id: quote for quote in quotes}
        _, temp_fasta = tempfile.mkstemp(suffix=".fa")
        with open(temp_fasta, "w+") as f:
            for quote in quotes:
                f.write(">%s\n%s\n" % (quote.id, quote.sequence))
        results = blast_sequence(self.sequence,
                                 subject=temp_fasta,
                                 word_size=10, perc_identity=100)

        if isinstance(results, list):
            alignments = sum([rec.alignments for rec in results], [])
        else:
            alignments = results.alignments

        for al in alignments:
            hit = max(al.hsps, key=lambda hit: hit.align_length)
            final_location = sorted((hit.query_start, hit.query_end))
            matching_segment = sorted((hit.sbjct_start, hit.sbjct_end))
            quotes_dict[al.hit_def].final_location = final_location
            quotes_dict[al.hit_def].matching_segment = matching_segment
        os.remove(temp_fasta)

    def tree_as_list(self):
        """Return a list containing the current AssemblyOperation and all its
        sub-operations and their respective sub-operations.

        Said otherwise, it flattens the assembly tree into the list of all
        nodes.
        """
        result = [self]
        if self.assembly_plan is not None:
            result += sum([
                child.tree_as_list()
                for segment, child in self.assembly_plan.items()
            ], [])
        return result

    def assembly_tree_as_dict(self, as_json=False, json_indent=None):
        """Return a JSON-like version of the nested tree.

        Parameters
        ----------

        as_json
          If True, a JSON string is returned, else the result is a dict object.

        json_indent
          number of spaces in the JSON indentation (for pretty printing). The
          default None means that the JSON will be on one line (TODO: check).



        Returns
        -------

          {
          "id": self.id,
          "source": self.source.name,
          "price": self.price,
          "lead_time": self.lead_time,
          "sequence": self.sequence,
          "message": self.message,
          "metadata" = self.metadata,
          "assembly_plan": { (start1, end1): {(subquote_1)},
                             (start2, end2): {(subquote_2)},
                           }
         }
        """
        final_location = (self.final_location if
                          hasattr(self, "final_location")
                          else None)
        matching_segment = (self.matching_segment if
                            hasattr(self, "matching_segment")
                            else None)

        assembly_plan = []
        if self.assembly_plan is not None:
            for (segment, quote) in self.assembly_plan.items():
                quote_as_dict = quote.assembly_tree_as_dict()
                quote_as_dict["segment_start"] = segment[0]
                quote_as_dict["segment_end"] = segment[1]
                assembly_plan.append(quote_as_dict)
        tree = {
            "id": self.id,
            "source": self.source.name,
            "price": self.price,
            "lead_time": self.lead_time,
            "sequence": self.sequence,
            "message": self.message,
            "metadata": self.metadata,
            "assembly_plan": assembly_plan,
            "final_location": final_location,
            "matching_segment": matching_segment,
            "accepted": self.accepted
        }

        if as_json:
            return json.dumps(tree, indent=json_indent)
        else:
            return tree


    def propagate_deadline(self, deadline):
        """Add a `deadline` attribute to the quote and propagate it to
        the quote's children by taking into account the duration of operations

        For instance if "self" has a duration of 5 and receives a deadline
        of 8, the quotes that "self" depends on will receive a deadline of
        8-5=3.
        """
        self.deadline = deadline
        children_deadline = deadline - self.step_duration
        if self.assembly_plan is not None:
            for segment, child in self.assembly_plan.items():
                child.propagate_deadline(children_deadline)

    @property
    def step_duration(self):
        """Duration of the final step of the assembly in the quote.

        Inferred by substracting the lead time of children operations to the
        lead time of this operation.
        """
        children_lead_time = self.children_overall_lead_time()
        if children_lead_time is None:
            return self.lead_time
        else:
            return self.lead_time - self.children_overall_lead_time()

    def compute_assembly_levels(self):
        """Return edges and levels for drawing the assembly graph nicely.

        """

        levels = defaultdict(lambda: [])
        edges = []

        def rec(subtree, depth=0):

            levels[depth].append(subtree)

            if subtree.assembly_plan is not None:
                for other in subtree.assembly_plan.values():
                    edges.append((other, subtree))
                    rec(other, depth + 1)
        rec(self)
        levels = [levels[i] for i in sorted(levels.keys())][::-1]
        levels = [sorted(level, key=lambda e: e.id)[::-1] for level in levels]
        return edges, levels

    def assembly_step_summary(self):
        """Return a print-friendly, simple string of the ordering plan."""
        plan = "\n  ".join(
            str(segment) + ": " + str(quote)
            for segment, quote in sorted(self.assembly_plan.items())
        )
        final_txt = "Ordering plan:\n  %s\n  Price:%.02f" % (plan, self.price)
        if self.lead_time is not None:
            final_txt = final_txt + ", total lead_time:%.1f" % self.lead_time
        return final_txt

    def children_total_price(self):
        """Return the total price of all sub-operations (apart from current)"""
        # print ([quote.accepted for quote in self.assembly_plan.values()])
        return sum(quote.price for quote in self.assembly_plan.values())

    def children_overall_lead_time(self):
        """Return the max lead time of all sub-operation (current one not
        included)"""
        if self.assembly_plan is None:
            return None
        lead_times = [quote.lead_time for quote in self.assembly_plan.values()]
        if len(lead_times) == 0:
            return None
        elif any(lead_time is None for lead_time in lead_times):
            return None
        else:
            return max(lead_times)

    def cuts_indices(self):
        """Return the locations of the `cuts` where the sequence is assembled.
        """
        return sorted([segment[1] for segment in self.assembly_plan])[:-1]

    def to_SeqRecord(self, record=None, record_id=None):
        """Return a Biopython seqrecord of the quote.

        >>> record = to_SeqRecord(solution)
        >>> # Let's plot with DnaVu:
        >>> from dnavu import create_record_plot
        >>> from bokeh.io import output_file, show
        >>> output_file("view.html")
        >>> plot = create_record_plot(record)
        >>> show(plot)
        """
        if record_id is None:
            record_id = self.id
        if record is None:
            record = SeqRecord(Seq(self.sequence, DNAAlphabet()), id=record_id)
        else:
            record = deepcopy(record)

        if self.assembly_plan is not None:
            features = [
                SeqFeature(
                    FeatureLocation(segment[0], segment[1], 1),
                    type="Feature",
                    qualifiers={
                        "name": quote.id,
                        "source": quote.source,
                        "price": quote.price,
                        "lead_time": quote.lead_time
                    }
                )
                for segment, quote in self.assembly_plan.items()
            ]
            record.features = features + record.features
        return record


    def to_genbank(self, filename=None, filehandle=None,
                   record=None, record_id=None):
        record = self.to_SeqRecord(record=record, record_id=record_id)
        if filename is not None:
            with open(filename, "w+") as f:
                SeqIO.write(record, f, "genbank")
        else:
            output = StringIO()
            SeqIO.write(record, output, "genbank")
            return output.getvalue()



from ..DnaQuote import DnaQuote
from ..SequenceDecomposer import SequenceDecomposer, NoSolutionFoundError
# from .shortest_path_algorithms import NoSolutionFoundError
from .DnaSource import DnaSource
from .DnaSourcesComparator import DnaSourcesComparator
from ..DnaAssemblyMethod import (BuildAGenomeAssemblyMethod,
                                 GibsonAssemblyMethod,
                                 GoldenGateAssemblyMethod)
from ..OverhangSelector import TmOverhangSelector, FixedSizeOverhangSelector
from ..constraints import SequenceLengthConstraint
import numpy as np

class DnaAssemblyStation(DnaSource):
    """DNA Assembly stations assemble together DNA fragments using a specific
    assembly method.

    Parameters
    ----------

    name
      Name of the station (appears on reports)

    assembly_method
      AnDnaAssemblyMethod object specifying how the fragments are assembled,
      what sequences can be assembled, what fragments can be used, etc.

    dna_source

    """
    class_description = "DNA assembly station"
    operation_type = "assembly"
    report_fa_symbol = u""
    report_fa_symbol_plain = "flask"
    report_color = "#eeeeff"

    def __init__(self, name, assembly_method, dna_source, memoize=False,
                 decomposer_class=None, a_star_auto_multiplier=2,
                 **solver_kwargs):
        self.name = name
        self.assembly_method = assembly_method
        self.set_suppliers(dna_source)
        self.extra_time = assembly_method.duration
        self.extra_cost = assembly_method.cost
        self.sequence_constraints = assembly_method.sequence_constraints
        self.cuts_set_constraints = assembly_method.cuts_set_constraints
        self.memoize = memoize
        if decomposer_class is None:
            self.decomposer_class = SequenceDecomposer
        else:
            self.decomposer_class = decomposer_class
        self.memoize_dict = {}
        self.min_basepair_price = self.dna_source.min_basepair_price
        if solver_kwargs.get('a_star_factor', None) == 'auto':
            solver_kwargs['a_star_factor'] = 2 * self.min_basepair_price
        self.solver_kwargs = solver_kwargs

    def get_quote_for_sequence_segment(self, sequence, segment,
                                       max_lead_time=None, **kwargs):
        """Return the cost of the segment

        Is used as the "cost" function for a segment during decomposition
        optimization

        """
        fragment_to_order = self.assembly_method.compute_sequence_fragment(
            sequence, segment, **kwargs
        )
        return self.dna_source.get_quote(fragment_to_order,
                                         max_lead_time=max_lead_time)

    def get_assembly_plan_from_cuts(self, sequence, cuts, max_lead_time=None):
        """Return a plan {(seg,ment): quote, ...} based on the cut positions.
        """
        cuts = sorted(cuts)
        return {
            segment: self.get_quote_for_sequence_segment(
                sequence, segment,
                max_lead_time=max_lead_time
            )
            for segment in zip(cuts, cuts[1:])
        }

    def new_sequence_decomposer(self, sequence, max_lead_time=None,
                                coarse_grain=None, fine_grain=None,
                                a_star_factor=0, logger=None):
        def segment_score(segment):
            quote = self.get_quote_for_sequence_segment(
                sequence, segment, max_lead_time=max_lead_time
            )
            if not quote.accepted:
                # print ("a")
                return -1
            else:
                return quote.price
        assembly = self.assembly_method
        if coarse_grain is None:
            coarse_grain = self.solver_kwargs.get("coarse_grain", 1)
        if fine_grain is None:
            fine_grain = self.solver_kwargs.get("fine_grain", 1)

        return self.decomposer_class(
            sequence_length=len(sequence),
            segment_score_function=segment_score,
            cut_location_constraints=[
                cs(sequence)
                for cs in assembly.cut_location_constraints
            ],
            segment_constraints=[
                cs(sequence)
                for cs in assembly.segment_constraints
            ],
            min_segment_length=assembly.min_segment_length,
            max_segment_length=assembly.max_segment_length,
            forced_cuts=assembly.force_cuts(sequence),
            suggested_cuts=self.compute_suggested_cuts(sequence),
            coarse_grain=coarse_grain,
            fine_grain=fine_grain,
            logger=logger,
            bar_prefix="[%s] " % self.name,
            a_star_factor=a_star_factor,
            path_size_limit=assembly.max_fragments,
            cuts_set_constraints=[
                cs(sequence)
                for cs in assembly.cuts_set_constraints
            ]
        )
    
    def compute_suggested_cuts(self, sequence):
        cuts = list(self.assembly_method.suggest_cuts(sequence))
        for supplier in self.suppliers:
            if hasattr(supplier, 'suggest_cuts'):
                cuts.extend(list(supplier.suggest_cuts(sequence)))
        return sorted(set(cuts))
        

    def get_assembly_plan_for_sequence(self, sequence, max_lead_time=None,
                                       coarse_grain=None, fine_grain=None,
                                       a_star_factor=0, logger=None):
        """Return the plan {(seg, ment): quote, ...} of the optimal strategy
        for the sequence's decomposition."""
        decomposer = self.new_sequence_decomposer(
            sequence, max_lead_time=max_lead_time, coarse_grain=coarse_grain,
            fine_grain=fine_grain, a_star_factor=a_star_factor, logger=logger)
        self._lattest_decomposer = decomposer  # for debugging
        best_cuts = decomposer.compute_optimal_cuts()
        return self.get_assembly_plan_from_cuts(sequence, best_cuts,
                                                max_lead_time=max_lead_time)

    def get_best_price(self, sequence, max_lead_time=None,
                       with_assembly_plan=False):
        """Returns a price-optimal DnaQuote for the given sequence.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quots

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
        """

        if (max_lead_time is not None):
            max_lead_time = max_lead_time - self.extra_time
            if max_lead_time < 0:
                return DnaQuote(self, sequence, accepted=False,
                                message="Lead time limit too short")
        try:
            assembly_plan = self.get_assembly_plan_for_sequence(
                sequence, max_lead_time=max_lead_time, **self.solver_kwargs
            )
        except NoSolutionFoundError:
            return DnaQuote(self, sequence, accepted=False,
                            message="No solution found !")

        # A solution has been found ! Now compute overall time and lead time.

        if any([q.price is None for q in assembly_plan.values()]):
            return DnaQuote(self, sequence, accepted=False,
                            message="No solution found !")
        quote = DnaQuote(self, sequence, assembly_plan=assembly_plan)
        quote.price = quote.children_total_price() + self.extra_cost
        children_lead_time = quote.children_overall_lead_time()
        if (children_lead_time is None) or (self.extra_time is None):
            quote.lead_time = None
        else:
            quote.lead_time = children_lead_time + self.extra_time
        if not with_assembly_plan:
            self.assembly_plan = None
        return quote

    def additional_dict_description(self):
        result = {
            "dna source": self.dna_source.name,
            "solver parameters": self.solver_kwargs,
        }
        result.update({
            ("assembly method %s" % k): v
            for k, v in self.assembly_method.dict_description().items()
        })
        return result

    @staticmethod
    def from_dict(data):
        sequence_constraints = []
        if data['use_size_range']:
            mini, maxi = data['size_range']
            sequence_constraints.append(SequenceLengthConstraint(mini, maxi))
        min_length, max_length = data['fragments_size_range']
        if data['method'] == 'type_iis':
            gc_range = data.get('overhang_gc_range', [0, 1])
            method = GoldenGateAssemblyMethod(
                duration=data['duration'],
                cost=data['cost'],
                min_overhangs_gc=gc_range[0],
                max_overhangs_gc=gc_range[1],
                enzyme=data['enzyme'],
                max_segment_length=max_length,
                min_segment_length=min_length,
                max_fragments=data['max_fragments'],
                sequence_constraints=sequence_constraints
            )
        elif data['method'] in ['gibson_assembly', 'yeast_recombination',
                                'oligo_assembly']:
            if data['overhang_type'] == 'tm':
                min_oh_size, max_oh_size = data['overhang_size_range']
                min_tm, max_tm = data['tm_range']
                overhang_selector = TmOverhangSelector(
                    min_size=min_oh_size, max_size=max_oh_size,
                    min_tm=min_tm, max_tm=max_tm
                )
            else:
                overhang_selector = FixedSizeOverhangSelector(
                    overhang_size=data['overlap']
                )
            if data['method'] == 'oligo_assembly':
                method_class = BuildAGenomeAssemblyMethod
            else:
                method_class = GibsonAssemblyMethod
            method = method_class(
                duration=data['duration'],
                cost=data['cost'],
                overhang_selector=overhang_selector,
                max_segment_length=max_length,
                min_segment_length=min_length,
                max_fragments=data['max_fragments'],
                sequence_constraints=sequence_constraints
            )
        if data["use_astar"]:
            a_star_factor = data.get('astar_factor', 'auto')
        else:
            a_star_factor = 0
        max_construct_length = max_length * data['max_fragments']
        if data['grain_type'] == 'auto':
            data['coarse_grain'] = int(max_construct_length / 15)
            data['fine_grain'] = int(np.sqrt(data['coarse_grain']))

        return DnaAssemblyStation(
            name=data['name'],
            dna_source=data['suppliers'],
            assembly_method=method,
            coarse_grain=data['coarse_grain'],
            fine_grain=data['fine_grain'],
            a_star_factor=a_star_factor,
            memoize=data.get('memoize', False),
            logger=data.get('logger', None)
        )

    def set_suppliers(self, suppliers):
        if hasattr(suppliers, '__iter__'):
            if len(suppliers) > 1:
                self.dna_source = DnaSourcesComparator(
                    name=self.name + ' comparator', suppliers=suppliers)
                self.dna_source.is_ghost_source = True
            else:
                self.dna_source = suppliers[0]
        else:
            self.dna_source = suppliers
            suppliers = [suppliers]
        self.suppliers = suppliers





import weasyprint
import matplotlib.pyplot as plt
from copy import deepcopy
import datetime
import os
import flametree

from jinja2 import Environment, FileSystemLoader
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Alphabet import DNAAlphabet
from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation

try:  # Python2 vs. Python3 ways of dealing with file-like strings
    from StringIO import StringIO
    USE_BYTES = False
except ImportError:
    from io import StringIO, BytesIO
    USE_BYTES = True

import pandas as pd

from .config import SETTINGS
from .plotting import (plot_assembly_blocks,
                       plot_supply_graph,
                       plot_assembly_graph,
                       give_quotes_html_locations,
                       ax_to_pdf)


pd.set_option('display.max_colwidth', -1)


def quote_to_SeqRecord(quote, record=None, record_id=None):
    """Return a Biopython seqrecord of the quote.

    >>> record = to_SeqRecord(solution)
    >>> # Let's plot with DnaVu:
    >>> from dnavu import create_record_plot
    >>> from bokeh.io import output_file, show
    >>> output_file("view.html")
    >>> plot = create_record_plot(record)
    >>> show(plot)
    """
    if record_id is None:
        record_id = quote.id
    if record is None:
        record = SeqRecord(Seq(quote.sequence, DNAAlphabet()), id=record_id)
    else:
        record = deepcopy(record)

    if quote.assembly_plan is not None:
        features = [
            SeqFeature(
                FeatureLocation(q.segment_start, q.segment_end, 1),
                type="misc_feature",
                qualifiers={
                    "label": "%s - From %s" % (q.id, q.source),
                    "name": q.id,
                    "source": q.source,
                    "price": q.price,
                    "lead_time": q.lead_time
                }
            )
            for q in quote.assembly_plan
        ]
        record.features = features + record.features
    return record


def quote_to_genbank(quote, filename=None, filehandle=None,
                     record=None, record_id=None):
    record = quote_to_SeqRecord(quote, record=record, record_id=record_id)
    if filename is not None:
        with open(filename, "w+") as f:
            SeqIO.write(record, f, "genbank")
    else:
        output = StringIO()
        SeqIO.write(record, output, "genbank")
        return output.getvalue()


def make_html_report(quote):
    """Return a HTML version of the assembly report (later converted to PDF)"""

    def _anchor_span(op):
        return "<span id='anchor_%s'>%s</span>" % (op.id, op.id)

    operations = quote.to_quotes_list()

    def sorting_key(op):
        location = op.get("final_location", None)
        if location is None:
            location = (-1, 1)
        return (location[0], -location[1])

    operations = sorted(operations, key=sorting_key)

    give_quotes_html_locations(
        operations, len_sequence=len(quote.tree.sequence))

    path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                        "templates")

    env = Environment(loader=FileSystemLoader(path))
    template = env.get_template("dnaweaver_jinja.html")

    orders_dataframe = pd.DataFrame.from_records([
        {
            "ID": _anchor_span(op),
            "Company": op.source,
            "Sequence": op.sequence,
            "Length": len(op.sequence),
            "Price": op.price,
            "Lead time": op.lead_time,
            "Deadline": op.get("deadline", None),
            "Location": op.html_location,
        }
        for op in operations
        if quote.sources[op.source].operation_type == "order"
    ])

    pcrs_dataframe = pd.DataFrame.from_records([
        {
            "ID": _anchor_span(op),
            "Primers": ", ".join([
                _anchor_span(child)
                for child in op.get("assembly_plan", [])
            ]),
            "Source": op.source,
            "Sequence": op.sequence,
            "Length": len(op.sequence),
            "Price": op.price,
            "Lead time": op.lead_time,
            "Deadline": op.get("deadline", None),
            "Location": op.html_location,
        }
        for op in operations
        if quote.sources[op.source].operation_type == "PCR"
    ],
        columns=["ID", "Source", "infos", "Primers",
                 "Sequence", "Length", "Price", "Lead time", "Location"]
    )

    parts_dataframe = pd.DataFrame.from_records([
        {
            "ID": _anchor_span(op),
            "Source": op.source,
            "Sequence": op.sequence,
            "Length": len(op.sequence),
            "Price": op.price,
            "Lead time": op.lead_time,
            "Deadline": op.get("deadline", None),
            "Location": op.html_location,
        }
        for op in operations
        if quote.sources[op.source].operation_type == "library"
    ],
        columns=["ID", "Source", "Sequence", "Length", "Price",
                 "Lead time", "Location"]
    )

    assembly_operations = [
        op
        for op in operations
        if quote.sources[op.source].operation_type == "assembly"
    ]

    asm_dataframe = pd.DataFrame.from_records([
        {
            "ID": _anchor_span(op),
            "Station": op.source,
            "Sequence": op.sequence,
            "Length": len(op.sequence),
            "Price": op.price,
            "Lead time": op.lead_time,
            "Deadline": op.get("deadline", None),
            "Location": op.html_location,
            "Primers": None
        }
        for op in assembly_operations
    ],
        columns=["ID", "Station", "infos", "Sequence",
                 "Length", "Price", "Lead time", "Location"]
    )

    cost_orders = orders_dataframe["Price"].sum()

    render_parameters = {
        "sequence": quote.tree.sequence,
        "sequence_name": quote.tree.get("sequence_name", "unknown"),
        "sequence_hash": "%02x" % hash(quote.tree.sequence),
        "quote": quote.tree,
        "date_submitted": datetime.datetime.now(),
        "n_orders": len(orders_dataframe),
        "cost_orders": cost_orders,
        "total_cost": quote.tree.price,
        "orders": orders_dataframe,
        "asm_dataframe": asm_dataframe,
        "assembly_operations": assembly_operations,
        "user_name": quote.tree.get("user_name", "unknown")
    }
    if len(pcrs_dataframe) > 0:
        render_parameters["pcr_reuses"] = pcrs_dataframe

    if len(parts_dataframe) > 0:
        render_parameters["parts_reuses"] = parts_dataframe

    html_out = template.render(render_parameters)

    return html_out


def make_pdf_report(quote, filename=None):
    """Return a PDF version of the report with general infos and details of
    each intermediary constructs."""
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)))
    base_path = os.path.join(path, "templates")
    css_path = os.path.join(base_path, "static", "css")
    stylesheets = [
        os.path.join(css_path, "dnaweaver_report.css"),
        # os.path.join(css_path, "font-awesome.min.css")
    ]
    html_out = make_html_report(quote)
    weasy_html = weasyprint.HTML(string=html_out, base_url=base_path)
    if filename is not None:
        write_mode = "wb" if USE_BYTES else "w+"
        with open(filename, write_mode) as f:
            weasy_html.write_pdf(f, stylesheets=stylesheets)
    else:
        output = BytesIO() if USE_BYTES else StringIO()
        weasy_html.write_pdf(output, stylesheets=stylesheets)
        return output.getvalue()


def _get_folder_readme_content():
    """General function to get the content of the FOLDER_README.txt template"""
    readme_path = os.path.join(SETTINGS["template_path"], "FOLDER_README.txt")
    with open(readme_path, "r") as f:
        readme_content = f.read()
    return readme_content


def make_spreadsheet_sequences_report(quote, filename=None):

    quotes_list = quote.to_quotes_list()
    sequences_dataframe = pd.DataFrame.from_records(
        [
            {
                "ID": _quote.id,
                "Type": _quote.get("operation_type", None),
                "Source": _quote.source,
                "SequenceLength": len(_quote.sequence),
                "Sequence": _quote.sequence
            }
            for _quote in quotes_list
        ],
        columns=["ID", "Type", "Source", "SequenceLength", "Sequence"]
    ).sort_values(by=["Type", "Source", "ID"])

    result = sequences_dataframe.to_csv(sep=";", index=False)
    if filename is not None:
        with open(filename, "w+") as f:
            f.write(result)
    else:
        return result


def make_folder_report(quote, target="@memory"):
    """Generate an extensive, multifile report as either a folder or a zip."""

    def write_ax_as_pdf(ax, target):
        ax.figure.savefig(target, format="pdf", bbox_inches="tight")

    root = flametree.file_tree(target, replace=True)
    root._file("assembly_report.pdf").write(
        make_pdf_report(quote))

    root._file('sequences.csv').write(
        make_spreadsheet_sequences_report(quote))

    root._file("README.txt").write(
        _get_folder_readme_content())

    # FIGURES

    figures = root._dir("figures")

    pos, ax = plot_supply_graph(quote)
    write_ax_as_pdf(ax, figures._file("supply_network.pdf").open("wb"))
    plt.close(ax.figure)

    pos, ax = plot_assembly_graph(quote, ax=None, textprops=None)
    write_ax_as_pdf(ax, figures._file("assembly_graph.pdf").open("wb"))
    plt.close(ax.figure)

    fig, ax_assembly_blocks = plt.subplots(1, figsize=(7, 4))
    assembly_blocks_ax, lg = plot_assembly_blocks(
        quote, backend="matplotlib", plot_top_assembly=False,
        ax=ax_assembly_blocks, parts_offset=0.1, legend=True)
    assembly_blocks_ax.figure.subplots_adjust(bottom=0.3)
    write_ax_as_pdf(assembly_blocks_ax,
                    figures._file("assembly_blocks.pdf").open("wb"))
    plt.close(fig)

    # GENBANKS

    genbank = root._dir("genbank")
    for q in quote.to_quotes_list():
        genbank._file(q.id + ".gb").write(quote_to_genbank(q))

    return root._close()



from .biotools import (reverse_complement, find_enzyme_sites,
                       gc_content_to_tm)
from .OverhangSelector import TmOverhangSelector
from .tools import memoize
import itertools


class DnaAssemblyMethod(object):
    """General class for assembly methods.

    All assembly methods can store the following attributes:

    Parameters
    ----------

    duration
      Duration required for the assembly (e.g. an estimated upper bound). The
      time unit is left at the choice of the user.

    cut_location_constraints
      List or tuple of functions `(sequence, int) -> bool` which return for
      a sequence and an index (cut location) whether the location should be
      considered as a cutting site compatible with this assembly method (True)
      or not.
      The locations considered in fine are the locations which pass every
      constraint in the `cut_location_contraints` list.

    segment_constraints
      List or tuple of functions `(sequence, start, end) -> bool` which returns
      for a subsegment `(start, end)` whether the segment is a valid segment
      for the assembly.
      The segments considered in fine are the segments which pass every
      constraint in the `segments_constraints` list.

    min_segment_length
      Minimal length of the fragments that this assembly method allows

    max_segment_length
      Maximal length of the fragments that this assembly method allows


    force_cuts
      A function `sequence->[i1, i2, i3...]` which for a given sequence returns
      forced cuts locations.

    max_fragments
       Maximal number of fragments which can be assembled with this method.

    sequence_constraints
      General constraints on the sequence so that it can be built with this
      assembly method. It is a list of functions `seq->bool`. If one of
      these constraints returns False the sequence is refused.

    reference
      Reference to e.g. a paper or a protocol describing the method.
    """
    name = "None"

    def __init__(self, duration=0, cost=0, reference=None,
                 cut_location_constraints=(),
                 segment_constraints=(), min_segment_length=0,
                 max_segment_length=None, force_cuts=(), suggest_cuts=(),
                 max_fragments=None, sequence_constraints=(),
                 cuts_set_constraints=()):
        self.duration = duration
        self.cost = cost
        self.cut_location_constraints = list(cut_location_constraints)
        self.segment_constraints = list(segment_constraints)
        self.min_segment_length = min_segment_length
        self.max_segment_length = max_segment_length
        self.sequence_constraints = list(sequence_constraints)
        self.max_fragments = max_fragments
        self.cuts_set_constraints = list(cuts_set_constraints)
        self.reference = reference

        if callable(suggest_cuts):
            self.suggest_cuts = suggest_cuts
        else:
            # means the cuts are a constant
            self.suggest_cuts = lambda *a: suggest_cuts

        if callable(force_cuts):
            self.force_cuts = force_cuts
        else:
            # means the forced cuts are a constant
            self.force_cuts = lambda *a: force_cuts

    def dict_description(self):
        result = {
            "name": self.name,
            "minimum segment length": self.min_segment_length,
            "maximum segment length": self.max_segment_length,
            "maximal number of fragments": self.max_fragments,
            "reference": self.reference,
            "duration": self.duration,
            "cost": self.cost
        }
        result.update(self.additional_dict_description())
        return result

    def additional_dict_description(self):
        return {}


class OverlapingAssemblyMethod(DnaAssemblyMethod):
    """General class for all overlaping assembly methods.

    Parameters
    ----------

    homology_arm_length
      Length of the homology arm, or "overhang". A length of L means that
      consecutive segments will overlap by 2*L

    """
    name = "Overlaping Assembly"

    def __init__(self, overhang_selector, **properties):
        super(OverlapingAssemblyMethod, self).__init__(**properties)
        selector = overhang_selector
        self.overhang_selector = selector
        if selector.has_location_filter:
            self.cut_location_constraints.append(
                selector.location_filter_method)
        self.compute_sequence_fragment = selector.compute_sequence_fragment


class GibsonAssemblyMethod(OverlapingAssemblyMethod):
    """Gibson Assembly Method. Just another overlap-method"""
    name = "Gibson Assembly"




class BuildAGenomeAssemblyMethod(OverlapingAssemblyMethod):
    """The Build-a-Genome Assembly Method. Just another overlap-method"""
    name = "Build-a-Genome"


class GoldenGateAssemblyMethod(OverlapingAssemblyMethod):
    """The Golden Gate Assembly Method.

    This method adds overhangs with a Type IIS REase site to the segments.

    Parameters
    ----------

    left_overhang
      An ATGC DNA sequence representing the left overhang, in 5' -3'.
      For practicality, this can contain "[BsaI]", "[BsmBI]", "[BbsI]", which
      will be replaced by their restriction sites


    right_overhang
      An ATGC DNA sequence representing the right overhang, in 5'-3'.
      Be careful, I said 5'-3' !!!
      If left to None, will be equal to left_overhang.

    """
    name = "Golden Gate Assembly"

    enzymes_dict = {
        "BsaI": "GGTCTC",
        "BsmBI": "CGTCTC",
        "BbsI": "GAAGAC",
    }

    def __init__(self, enzyme="BsaI", wildcard_basepair="A",  left_addition="",
                 right_addition="", refuse_sequences_with_enzyme_site=True,
                 min_overhangs_gc=0, max_overhangs_gc=1,
                 min_overhangs_differences=1,  **props):
        if enzyme not in self.enzymes_dict:
            return ValueError("Enzyme should be one of %s" %
                              self.enzymes_dict.keys())

        self.min_overhangs_gc = min_gc = min_overhangs_gc
        self.max_overhangs_gc = max_gc = max_overhangs_gc
        self.min_overhangs_differences = min_overhangs_differences

        self.enzyme = enzyme
        self.enzyme_site = self.enzymes_dict[enzyme]
        enzyme_site_plus_basepair = self.enzyme_site + wildcard_basepair
        self.left_addition = left_addition + enzyme_site_plus_basepair
        self.right_addition = (reverse_complement(enzyme_site_plus_basepair) +
                               right_addition)
        self.refuse_sequences_with_enzyme_site = \
            refuse_sequences_with_enzyme_site

        overhang_selector = TmOverhangSelector(
            min_size=4, max_size=4,
            min_tm=gc_content_to_tm(4, min_gc),
            max_tm=gc_content_to_tm(4, max_gc),
            left_addition=self.left_addition,
            right_addition=self.right_addition
        )
        OverlapingAssemblyMethod.__init__(self, overhang_selector, **props)

        # CUTS LOCATION CONSTRAINT BASED ON GC CONTENT

        if refuse_sequences_with_enzyme_site:
            def no_site_in_sequence(sequence):
                sites = find_enzyme_sites(sequence, enzyme_name=self.enzyme)
                return sites == []
            self.sequence_constraints.append(no_site_in_sequence)

        # CUTS LOCATION CONSTRAINT BASED ON GC CONTENT



        # CUTS SET CONSTRAINT: ALL OVERHANGS MUST BE COMPATIBLE

        def overhangs_are_compatible(o1, o2):
            diffs = len([a for a, b in zip(o1, o2) if a != b])
            if diffs >= self.min_overhangs_differences:
                rev_o2 = reverse_complement(o2)
                rev_diffs = len([a for a, b in zip(o1, rev_o2) if a != b])
                return rev_diffs >= self.min_overhangs_differences
            return False

        overhangs_are_compatible = memoize(overhangs_are_compatible)

        def all_overhangs_are_compatible(sequence):
            def f(cut_locations):
                overhangs = sorted([
                    overhang_selector.compute_overhang_sequence(sequence,
                                                                cut_location)
                    for cut_location in cut_locations
                ])
                return all([
                    overhangs_are_compatible(o1, o2)
                    for o1, o2 in itertools.combinations(overhangs, 2)
                ])
            return f

        self.cuts_set_constraints.append(all_overhangs_are_compatible)

    def additional_dict_description(self):
        return {
            "enzyme": self.enzyme,
            "left addition": self.left_addition,
            "right addition": self.right_addition,
            "refuse sequences with enzyme site":
                str(self.refuse_sequences_with_enzyme_site),
            "overhangs gc content": "%d-%d%%" % (100*self.min_overhangs_gc,
                                                 100*self.max_overhangs_gc),
            "overhangs differences": self.min_overhangs_differences
        }



from ..DnaQuote import DnaQuote
from .DnaSource import DnaSource

class DnaSourcesComparator(DnaSource):
    """Special source that compares quotes from other DnaSources.

    Upon receiving a sequence, that source will submit the sequence to
    Downstream sources, which deliver each one optimal quote. The comparator
    then returns the one quote with the lowest price.

    Parameters
    ----------

    suppliers
      List of `DnaSources` that

    memoize
      Whether the quotes should be kept in memory to avoid re-computing the
      same quote several times. Can accelerate computations but is
      RAM-expensive.
    """
    class_description = "DNA sources comparator"
    operation_type = "comparison"
    report_fa_symbol = u""
    report_fa_symbol_plain = u"circle-o"
    report_color = "#000000"

    def __init__(self, suppliers=(), memoize=False, sequence_constraints=(),
                 name="comparator"):
        self.set_suppliers(suppliers)
        self.memoize = memoize
        self.sequence_constraints = sequence_constraints
        self.memoize_dict = {}
        self.name = name

    def get_best_price(self, sequence, max_lead_time=None,
                       with_assembly_plan=False):
        """Returns a price-optimal DnaQuote for the given sequence.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quots

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
        """

        best_quote = None
        best_basepair_price = 1e8
        for source in sorted(self.suppliers,
                             key=lambda s: s.min_basepair_price):
            # print (self.name, len(sequence), source)
            if source.min_basepair_price > best_basepair_price:
                break
            quote = source.get_quote(sequence, max_lead_time=max_lead_time,
                                     with_assembly_plan=with_assembly_plan)
            # print ('=>', quote.accepted, quote.price)
            if not quote.accepted:
                continue
            quote_basepair_price = quote.price / float(len(sequence))
            if quote_basepair_price < best_basepair_price:
                best_quote = quote
                best_basepair_price = quote_basepair_price
        # print ('---- went for ', best_quote.source)
        if best_quote is None:
           return DnaQuote(self, sequence, accepted=False,
                           message="Sequence was rejected by all sources.")
        return best_quote

    def additional_dict_description(self):
        return {
            "dna sources": [source.name for source in self.suppliers],
        }
    @staticmethod
    def from_dict(data):
        return DnaSourcesComparator(
             name=data['name'], suppliers=data['suppliers'])

    def set_suppliers(self, suppliers):
        self.suppliers = suppliers
        if len(suppliers):
            self.min_basepair_price = min([supplier.min_basepair_price
                                           for supplier in suppliers])
        else:
            self.min_basepair_price = None
    
    def suggest_cuts(self, sequence):
      return sorted(set([
        cut
        for supplier in self.suppliers
        if hasattr(supplier, 'suggest_cuts')
        for cut in supplier.suggest_cuts(sequence)
      ]))



"""DnaWeaver Reports implements reporting methods for Dna Weaver outputs."""

from .DnaSource import DnaSource
from .CommercialDnaOffer import CommercialDnaOffer
from .DnaAssemblyStation import DnaAssemblyStation
from .DnaSourcesComparator import DnaSourcesComparator
from .PartsLibrary import PartsLibrary, GoldenGatePartsLibrary
from .PcrOutStation import PcrOutStation



from copy import deepcopy
import numpy as np
import dnachisel as dc

def compute_cost_profiles(graph):
    
    L = max(graph.nodes)
    cost_profiles = {
        'min': 1000 * np.ones(L),
        'max': -1000 * np.ones(L),
        'mean': np.zeros(L),
        'total': np.zeros(L),
        'count': np.zeros(L)
    }
    for start, end, data in graph.edges(data=True):
        cost_profiles['count'][start:end] += 1.0
        bp_price = data['weight'] / (end - start)
        cost_profiles['total'][start:end] += bp_price
        cost_profiles['min'][start:end] = np.minimum(
            bp_price, cost_profiles['min'][start:end])
        cost_profiles['max'][start:end] = np.maximum(
            bp_price, cost_profiles['max'][start:end])
    cost_profiles['mean'] = cost_profiles['total'] / (
        .0001 + cost_profiles['count'])
    diffs = np.diff(cost_profiles['min'])
    cost_profiles['diffs'] = diffs
    return cost_profiles

class OptimizeManufacturability(dc.Specification):
    
    def __init__(self, station, detection_threshold=0.0005,
                 max_loc_size=50, resolution=10, boost=1):
        co_station = deepcopy(station)
        co_station.assembly_method.min_segment_length = resolution
        co_station.a_star_factor = 0
        self.station = station
        self.co_station = co_station
        self.resolution = resolution
        self.detection_threshold = detection_threshold
        self.boost = boost
        self.max_loc_size = max_loc_size
        self.is_localized = False
    
    def compute_cost_graph(self, sequence):
        decomposer = self.co_station.new_sequence_decomposer(sequence)
        grained_cuts = set(range(0, decomposer.sequence_length + 1,
                           self.resolution))
        grained_cuts = grained_cuts.union(set([decomposer.sequence_length]))
        graph = decomposer.compute_graph(grained_cuts,
                                         reachable_indices_only=False)
        graph, cuts = decomposer.find_shortest_path(graph)
        return graph
    
    def compute_cost_profiles(self, sequence):
        graph = self.compute_cost_graph(sequence)
        return compute_cost_profiles(graph)

    def detect_cost_driving_locations(self, sequence):
        diffs = self.compute_cost_profiles(sequence)['diffs']
        above_threshold = abs(diffs) > self.detection_threshold
        changes = [
            (index, diffs[index])
            for index in above_threshold.nonzero()[0]
        ]
        return [
            dc.Location(start[0], end[0])
            for start, end in zip(changes, changes[1:])
            if start[1] > 0
            and end[1] < 0
            and end[0] - start[0] < self.max_loc_size
        ]
        
    def evaluate(self, problem):
        if not self.is_localized:
            locations = self.detect_cost_driving_locations(problem.sequence)
        else:
            locations = []
        quote = self.get_quote(problem)
        score = -quote.price
        return dc.SpecEvaluation(self, problem=problem, score=score,
                                 locations=locations, data=quote)

    def get_quote(self, problem):
        return self.station.get_quote(problem.sequence,
                                      with_assembly_plan=False)

    def localized(self, location, problem=None):
        return self.copy_with_changes(is_localized=True)


# -*- coding: utf-8 -*-

from collections import defaultdict

import numpy as np
from ..DnaQuote import DnaQuote
# Attempt import of optional module DNA Chisel
try:
    from dnachisel import Specification, DnaOptimizationProblem
    DNACHISEL_AVAILABLE = True
except:
    Specification = type("BLANK")  # meant to be a fake type that matches nothing
    DNACHISEL_AVAILABLE = False



class DnaSource:
    """Base class for all DnaSources, which are the elements of the supply
    networks used to define assembly problems in DnaWeaver."""

    min_basepair_price = 0

    def get_quote(self, sequence, max_lead_time=None, max_price=None,
                  with_assembly_plan=False, time_resolution=1.0):
        """Return a DnaQuote with price, lead time, etc. for a given sequence.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quots

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        max_price (float)
          If provided, the quote returned is the least-lead-time quote
          whose price is below or equal to `max_price`.
          This is done using bisection and can be slow as it requires to
          re-compute the problem many times
          Note that either this parameter or `max_lead_time` must be None

        with_assembly_plan
          If True, the assembly plan is added to the quote

        time_resolution
          Time resolution for the bisecting search if `max_price` is not None.

        Returns
        -------

        A DnaQuote object.

        """

        if max_lead_time is None:
            max_lead_time = np.inf

        if self.memoize:
            args = (hash(sequence), max_lead_time, max_price,
                    with_assembly_plan)
            quote = self.memoize_dict.get(args, None)
            if quote is not None:
                return quote

        if not self.verify_constraints(sequence):
            quote = DnaQuote(self, sequence, accepted=False,
                             message="Sequence does not pass constraints.")

        elif max_price is not None:
            quote = self.get_best_lead_time_under_price_limit(
                sequence,
                max_price=max_price,
                time_resolution=time_resolution,
                with_assembly_plan=with_assembly_plan,

            )
        else:
            quote = self.get_best_price(
                sequence,
                max_lead_time=max_lead_time,
                with_assembly_plan=with_assembly_plan
            )

        if self.memoize:
            self.memoize_dict[args] = quote

        return quote

    def __repr__(self):
        return self.name

    def get_best_lead_time_under_price_limit(self, sequence, max_price,
                                             time_resolution,
                                             with_assembly_plan=False):
        """Return the quote with fastest lead time under the budget constraint

        Parameters
        ----------



        sequence (str)
          The sequence submitted to the Dna Source for a quots

        max_price (float)
          If provided, the quote returned is the least-lead-time quote
          whose price is below or equal to `max_price`.
          This is done using bisection and can be slow as it requires to
          re-compute the problem many times
          Note that either this parameter or `max_lead_time` must be None

        with_assembly_plan
          If True, the assembly plan is added to the quote

        time_resolution
          Time resolution for the bisecting search if `max_price` is not None.

        """
        def f(max_lead_time):
            return self.get_quote(sequence, max_lead_time=max_lead_time,
                                  with_assembly_plan=with_assembly_plan)
        quote = f(None)
        if (not quote.accepted) or (quote.price > max_price):
            return DnaQuote(
                self, sequence, accepted=False,
                message="Price over limit even without time limit."
            )
        step_size = quote.lead_time / 2.0
        lead_time = quote.lead_time - step_size
        best_quote = quote
        while step_size > time_resolution:
            quote = f(lead_time)
            if quote.accepted and (quote.price <= max_price):
                best_quote = quote
                lead_time -= step_size
            else:
                lead_time += step_size
            step_size /= 2.0

        return best_quote

    def verify_constraints(self, sequence):
        """Return True iff `sequence` passes all `self.sequence_constraints`

        Will automatically process DNA-Chisel constraints that would be in
        `self.sequence_constraints`

        """
        constraints = self.sequence_constraints
        if not hasattr(self, 'dnachisel_constraints'):
            self.dnachisel_constraints = [
                constraint for constraint in self.sequence_constraints
                if isinstance(constraint, Specification)
            ]

        if self.dnachisel_constraints != []:
            if not DNACHISEL_AVAILABLE:
                raise ImportError("Spotted DNA Chisel constraints, while "
                                  "DNA Chisel is not installed.")
            # We provide an empty mutation space so it won't be recomputed
            # (which would take time and is useless here!)
            problem = DnaOptimizationProblem(sequence, dnachisel_constraints,
                                             mutation_space=[])
            constraints = [
                constraint for constraint in constraints
                if not isinstance(constraint, Specification)
            ] + [lambda seq: problem.all_constraints_pass()]

        return all(
            constraint(sequence)
            for constraint
            in constraints
        )

    def compute_supply_graph(self):
        """Return elements to plot the supply graph underlying this DnaSource

        Returns
        -------

        edges
          A list [(s1,s2), (s1,s3), (s2, s5)...] of couples of DnaSources in a
          supplier-supplied relationship.

        levels
          A list of lists [[s1,s2], [s4,s8,s9]...] of sources. The first
          sublist (first level) are all sources at the farthest distance from
          the current source in the supply graph, and the last sublist contains
          only the current DnaSource.
        """

        seen_sources = set()
        levels = defaultdict(lambda: [])
        edges = []

        def rec(source, depth=0):

            if source in seen_sources:
                return
            seen_sources.add(source)

            levels[depth].append(source)
            if hasattr(source, "suppliers"):
                for other in source.suppliers:
                    edges.append((other, source))
                    rec(other, depth + 1)
            elif hasattr(source, "dna_source"):
                edges.append((source.dna_source, source))
                rec(source.dna_source, depth + 1)
            elif hasattr(source, "primers_dna_source"):
                edges.append((source.primers_dna_source, source))
                rec(source.primers_dna_source, depth + 1)
            

        rec(self)

        levels = [levels[i] for i in sorted(levels.keys())][::-1]
        return edges, levels
    
    def prepare_network_on_sequence(self, sequence):
        edges, levels = self.compute_supply_graph()
        for level in levels:
            for source in level:
                if hasattr(source, 'prepare_on_sequence'):
                    source.prepare_on_sequence(sequence)

    def dict_supply_graph(self):
        sources = {}

        def rec(source, depth=0):

            if source in sources:
                return
            if hasattr(source, 'is_ghost_source'):
                return
            sources[source.name] = source.dict_description()
            sources[source.name]["_depth"] = depth

            providers = sources[source.name]["providers"] = []
            if hasattr(source, "suppliers"):
                for other in source.suppliers:
                    providers.append(other.name)
                    rec(other, depth + 1)
            elif hasattr(source, "dna_source"):
                providers.append(source.dna_source.name)
                rec(source.dna_source, depth + 1)
            elif hasattr(source, "primers_dna_source"):
                providers.append(source.primers_dna_source.name)
                rec(source.primers_dna_source, depth + 1)
            if hasattr(source, "dna_sources"):
                for other in source.dna_sources:
                    providers.append(other.name)
                    rec(other, depth + 1)
            
        rec(self)
        return sources

    def dict_description(self):
        result = {
            "name": self.name,
            "operation_type": self.operation_type,
            "class": self.class_description,
            "_report_color": self.report_color,
            "_report_fa_symbol": self.report_fa_symbol,
            "_report_fa_symbol_plain": self.report_fa_symbol_plain
        }
        result.update(self.additional_dict_description())
        return result

    def additional_dict_description(self):
        return {}


class FragmentAmplificationStation(DnaSource):
    """PCR-Out a fragment from a vector to linearize it for use in subsequent
    assemblies such as Gibson assembly."""
    report_fa_symbol = u""
    report_fa_symbol_plain = "exchange"
    report_color = "#eefefe"
    operation_type = "PCR"

    def __init__(self, fragment_dna_source,  primers_dna_source,
                 primer_melting_temperature=50, sequence_constraints=()):
        self.fragment_dna_source = fragment_dna_source
        self.primers_dna_source = primers_dna_source
        self.primer_melting_temperature = primer_melting_temperature
        self.sequence_constraints = sequence_constraints

    def additional_dict_description(self):
        return {
            "primers DNA source": self.primers_dna_source.name,
            "primers melting temp.": self.primer_melting_temperature
        }



# This will try to import setuptools. If not here, it will reach for the
# embedded ez_setup (or the ez_setup package). If none, it fails with a message
try:
    from setuptools import setup
except ImportError:
    try:
        import ez_setup
        ez_setup.use_setuptools()
    except ImportError:
        raise ImportError("DnaWeaver could not be installed, probably because"
                          " neither setuptools nor ez_setup are installed on"
                          "this computer. \nInstall ez_setup "
                          "([sudo] pip install ez_setup) and try again.")

from setuptools import setup, find_packages

exec(open('dnaweaver/version.py').read())  # loads __version__

setup(name='DnaWeaver',
      version=__version__,
      author='Zulko',
      description='Make ordering and assembly plans for DNA sequences',
      long_description=open('README.rst').read(),
      license='MIT',
      keywords="DNA optimization assembly ordering synthetic biology",
      packages=find_packages(exclude='docs'),
      include_package_data=True,
      install_requires=["numpy", "Biopython", 'proglog', "networkx",
                        "flametree"])



"""Optimization techniques"""

import numpy as np
import networkx as nx
from heapq import heappush, heappop
from itertools import count
from copy import deepcopy


class NoSolutionFoundError(Exception):
    """Error thrown when an optimization problem seems to have no solution."""
    pass


def astar_path(G, source, target, heuristic=None, weight=None):
    """Return a list of nodes, A* shortest path between source and target.

    Uses the A* ("A-star") algorithm.

    There may be more than one shortest path.  This returns only one.

    This function is taken from the Networkx project, with modifications for
    lazy weight computations (weight can now be a function).
    This function is therefore placed under the BSD licence.

    TO DO: submit that change to the NetworkX project so that this rewrite
    can be avoided.


    Parameters
    ----------

    G : NetworkX graph
    source : node
       Starting node for path
    target : node
       Ending node for path
    heuristic : function
       A function to evaluate the estimate of the distance
       from the a node to the target.  The function takes
       two nodes arguments and must return a number.
    weight: string, optional (default='weight')
       Edge data key corresponding to the edge weight.

    Raises
    ------

    NetworkXNoPath
        If no path exists between source and target.

    Examples
    --------
    >>> G=nx.path_graph(5)
    >>> print(nx.astar_path(G,0,4))
    [0, 1, 2, 3, 4]
    >>> G=nx.grid_graph(dim=[3,3])  # nodes are two-tuples (x,y)
    >>> def dist(a, b):
    ...    (x1, y1) = a
    ...    (x2, y2) = b
    ...    return ((x1 - x2) ** 2 + (y1 - y2) ** 2) ** 0.5
    >>> print(nx.astar_path(G,(0,0),(2,2),dist))
    [(0, 0), (0, 1), (1, 1), (1, 2), (2, 2)]

    See Also
    --------
    shortest_path, dijkstra_path
    """
    if source not in G or target not in G:
        msg = 'Either source {} or target {} is not in G'
        raise nx.exception.NetworkXNoPath(msg.format(source, target))

    if heuristic is None:
        # The default heuristic is h=0 - same as Dijkstra's algorithm
        def heuristic(u, v):
            return 0
    if weight is None:
        weight = lambda *a: 1
    elif isinstance(weight, str):
        key = weight
        def weight(n1, n2, props):
            return props[key]

    push = heappush
    pop = heappop
    c = count()
    queue = [(0, next(c), source, 0, None)]
    enqueued = {}
    explored = {}

    while queue:
        _, __, curnode, dist, parent = pop(queue)

        if curnode == target:
            path = [curnode]
            node = parent
            while node is not None:
                path.append(node)
                node = explored[node]
            path.reverse()
            return path

        if curnode in explored:
            continue

        explored[curnode] = parent

        for neighbor, w in G[curnode].items():
            if neighbor in explored:
                continue
            ncost = dist + weight(curnode, neighbor, w)
            if neighbor in enqueued:
                qcost, h = enqueued[neighbor]
                if qcost <= ncost:
                    continue
            else:
                h = heuristic(neighbor, target)
            enqueued[neighbor] = ncost, h
            push(queue, (ncost + h, next(c), neighbor, ncost, curnode))

    raise nx.NetworkXNoPath("Node %s not reachable from %s" % (source, target))


def shortest_valid_path(graph, start, end, nodes_constraints=(),
                        size_limit=None, min_step=None,
                        compatibility_search_cutoff=100,
                        penalty=0):
    if size_limit is not None:
        def f(penalty):
            try:
                path = shortest_valid_path(
                    graph, start, end, penalty=penalty, size_limit=None,
                    nodes_constraints=nodes_constraints,
                    compatibility_search_cutoff=100
                )
                return path, len(path) - 1
            except NoSolutionFoundError:
                return None, np.inf

        weights = [data["weight"] for _, _, data in graph.edges(data=True)]
        if weights == []:
            raise NoSolutionFoundError()
        penalty = max(weights)
        step = -0.5 * penalty

        best_path, best_path_size = f(penalty)
        if best_path_size > size_limit:
            raise NoSolutionFoundError()

        while abs(step) > min_step:
            penalty = penalty + step
            path, path_size = f(penalty)
            if path_size == size_limit:
                return path
            elif path_size > size_limit:
                step = 0.5 * abs(step)
            else:  # path_size < size_limit
                step = -0.5 * abs(step)
                best_path = path
        return best_path

    elif penalty > 0:
        _, _, edge_datas = zip(*graph.edges(data=True))
        for data in edge_datas:
            data["weight"] += penalty
        try:
            path = shortest_valid_path(
                graph, start, end, penalty=0, size_limit=None,
                compatibility_search_cutoff=compatibility_search_cutoff,
                nodes_constraints=nodes_constraints
            )
            for data in edge_datas:
                data["weight"] -= penalty
        except nx.NetworkXNoPath:
            for data in edge_datas:
                data["weight"] -= penalty
            raise NoSolutionFoundError()
        return path

    elif len(nodes_constraints) == 0:
        return nx.dijkstra_path(graph, start, end, weight='weight')

    else:
        def path_is_valid(path):
            return all(nodes_constraint(path)
                       for nodes_constraint in nodes_constraints)
        tentative_shortest_path = nx.dijkstra_path(graph, start, end,
                                                   weight='weight')
        if path_is_valid(tentative_shortest_path):
            return tentative_shortest_path
        shortest_paths = nx.shortest_simple_paths(graph, start, end,
                                                  weight='weight')
        for i in range(compatibility_search_cutoff):
            shortest_path = next(shortest_paths)
            if path_is_valid(shortest_path):
                return shortest_path
        return nx.NetworkXNoPath("Could not find a solution verifying the cuts"
                                 " set constraints, after %d tries." %
                                 compatibility_search_cutoff)



"""Plotting functions for the results of DNA Weaver"""

import textwrap
from copy import deepcopy
import itertools
from collections import defaultdict

import colorsys

from Bio import SeqIO

try:
    from StringIO import StringIO
    USE_BYTES = False
except ImportError:
    from io import StringIO, BytesIO
    USE_BYTES = True

from base64 import b64encode

try:
    from bokeh.plotting import figure, ColumnDataSource
    from bokeh.models import (Range1d, HoverTool, Legend)
    BOKEH_AVAILABLE = True
except ImportError:
    BOKEH_AVAILABLE = False
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as cl
import matplotlib.cm as cm
import matplotlib.font_manager as fm
from matplotlib.path import Path
import matplotlib.patches as mpatches
import matplotlib.gridspec as gridspec

from dna_features_viewer import GraphicRecord

from .config import SETTINGS


def plot_assembly_blocks(quote, parts_offset=0, backend="matplotlib",
                         plot_top_assembly=True,
                         ax=None, edge_widths=None, legend=False):
    """Return a Matplotlib or Bokeh plot of the assembly tree of blocks.

    Parameters
    ----------



    parts_offset
      Offset applied so that consecutive blocks are not exactly on the same
      level. Can go from 0 (flat line of blocks) to e.g. 1

    backend
      One of "matplotlib" or "bokeh". Determines the library used to produce
      the graph. matplotlib can output PNG/SVG/etc. while Bokeh outputs an
      interactive HTML/JS graph

    plot_top_assembly
      Whether the top assembly (which is just one big block) should be plotted
      or not

    ax
      A Matplotlib Axes object. If no ax is provided, a new figure and ax are
      generated.

    edge_widths
      A dict {sourcename : width} indicating the widths of the rectangles based
      on the source name. Rectangles with a very small width will be edgeless.

    legend
      Whether the legend is included in the ax.

    """
    rectangles = []
    if edge_widths is None:
        edge_widths = {}
    if not hasattr(quote, 'tree'):
        print('lol')
        quote.compute_full_assembly_tree()
    tree = deepcopy(quote.tree)

    def rec(_quote, depth=0, xstart=0, xend=None):
        children = _quote.pop("assembly_plan")
        #print _quote.final_location
        if _quote.final_location is not None:
            left, right = _quote.final_location
        else:
            left, right = xstart, xend

        if children is None:
            children = []
        if plot_top_assembly or (depth > 0):
            source = quote.sources[_quote.source]
            rectangles.append({
                "top": -depth + .5 + (not plot_top_assembly),
                "bottom": -depth + (not plot_top_assembly),
                "left": left,
                "right": right,
                "source": str(_quote.source),
                "lead_time": str(_quote.lead_time),
                "price": str(_quote.price),
                "length": len(_quote.sequence),
                "has_children": children != [],
                "line_width": 1,
                "operation_type": source.operation_type,
                "fa_symbol": source._report_fa_symbol_plain,
                "color": source._report_color
            })
        for child in children:
            rec(child, depth=depth + 1,
                xstart=xstart + child.segment_start,
                xend=xstart + child.segment_end)
    rec(tree, xend=len(tree.sequence))
    tops = list(set([r["top"] for r in rectangles]))
    offsets = {top: 0 for top in tops}

    guides = []
    for top in sorted(tops):
        rects = sorted([r for r in rectangles if r["top"] == top],
                       key=lambda r: r["left"])
        for i, rect in enumerate(rects):
            color = ["#e5ecff", "#ffffe5"][i % 2]
            if rect["has_children"]:
                guides.append({
                    "top": rect["top"],
                    "bottom": rect["top"] - 1,
                    "left": rect["left"],
                    "right": rect["right"],
                    "color": color
                })
    for rectangle in sorted(rectangles, key=lambda r: r["left"]):
        if parts_offset:
            offset = offsets[rectangle["top"]]
            offsets[rectangle["top"]] = (offsets[rectangle["top"]] + 1) % 2
            rectangle["top"] += parts_offset * offset
            rectangle["bottom"] += parts_offset * offset

    if backend == "bokeh":
        return _bokeh_plot_assembly_blocks(rectangles, guides, tops)
    elif backend == "matplotlib":
        if hasattr(quote, "genbank") and (quote.genbank is not None):
            record = SeqIO.read(quote.genbank, "genbank")
        else:
            record = None
        return _matplotlib_plot_assembly_blocks(
            rectangles, guides, tops, ax, record=record, legend=legend)
    else:
        raise ValueError("Backend should be bokeh or matplotlib")


def _bokeh_plot_assembly_blocks(rectangles, guides, tops):
    """Plot the block rectangles in a bokeh plot"""
    hover = HoverTool(
        tooltips="""
        <div>
          <i class="fa fa-@fa_symbol"></i>
          <b>@operation_type from @source </b><br/>
          @length bp, @price $, @lead_time days
          </div>
    """,
        names=["parts"]
    )

    L = max([r["right"] for r in rectangles])
    p = figure(tools=[hover, "xwheel_zoom,xpan,reset"],
               responsive=True,
               plot_height=300,
               x_range=Range1d(int(-0.05 * L), int(1.05 * L)),
               y_range=Range1d(min(tops) - 1, 1),
               logo=None, toolbar_location="above",
               title="Assembly Plan")
    source_guide = ColumnDataSource(pd.DataFrame.from_records(guides))
    p.quad(name="guides",
           top="top", bottom="bottom", left="left",
           right="right", color="color", source=source_guide, alpha=0.5)
    dataframe = pd.DataFrame.from_records(rectangles)
    legend_items = []
    for sourcename, df in dataframe.groupby("source"):
        source = ColumnDataSource(df)
        quads = p.quad(name="parts", top="top", bottom="bottom", left="left",
                       right="right", color="color", source=source,
                       line_width=1, line_color="black", alpha=1)
        legend_items.append([sourcename, [quads]])

    p.yaxis.visible = False
    p.outline_line_color = None
    p.grid.grid_line_color = None
    p.add_layout(Legend(items=legend_items, location=(0, -30)), 'right')
    return p


def _matplotlib_plot_assembly_blocks(rectangles, guides, tops, ax=None,
                                     fig_height="auto", fig_width=8,
                                     legend=False, textprops=None,
                                     record=None):
    """Plot the assembly block rectangles using matplotlib."""

    if record is not None:
        gs = gridspec.GridSpec(4, 1)
        fig = plt.figure(figsize=(7,7))
        record_ax = fig.add_subplot(gs[0])
        blocks_ax = fig.add_subplot(gs[1:])
        grecord = GraphicRecord.from_biopython_record(
            record, fun_color=lambda a: "#aabbff"
        )
        grecord.plot(record_ax)
        res = plot_assembly_blocks(rectangles, guides, tops, ax=blocks_ax,
                                   fig_height="auto", fig_width=8,
                                   legend=False, textprops=None,
                                   record=None)
        record_ax.set_xlim(blocks_ax.get_xlim())
        return res

    L = max([r["right"] for r in rectangles])

    if textprops is None:
        textprops = fm.FontProperties(
            fname=SETTINGS["OpenSans-ttf-path"],
            size=12, family="sans-serif"
        )

    if ax is None:
        if fig_height == "auto":
            fig_height = 8
        fig, ax = plt.subplots(1, figsize=(8, 8))
    ax.set_xlim((-0.05 * L, 1.05 * L))
    ax.set_ylim((min(tops) - 1, 1.5))

    for g in guides:
        patch = mpatches.Rectangle((g["left"], g["bottom"]),
                                   g["right"] - g["left"],
                                   g["top"] - g["bottom"],
                                   color=g["color"],
                                   ec="none")
        ax.add_patch(patch)

    seen_sources = set([])
    legend_handles = []
    sorted_rectangles = sorted(rectangles,
                               key=lambda r: (-r["bottom"], r["left"]))
    for g in sorted_rectangles:
        if g["source"] not in seen_sources:
            legend_patch = mpatches.Patch(facecolor=g["color"],
                                          label=g["source"],
                                          edgecolor="k",
                                          linewidth=1.0)
            legend_handles.append(legend_patch)
            seen_sources.add(g["source"])
            plt.legend(handles=[legend_patch])
        # linewidth = g["line_width"]
        width = g["right"] - g["left"]
        line_width = 1.0 if (1.0 * width / L) > .002 else 0
        patch = mpatches.Rectangle((g["left"], g["bottom"]),
                                   width,
                                   g["top"] - g["bottom"],
                                   facecolor=g["color"],
                                   edgecolor="k",
                                   linewidth=line_width)
        ax.add_patch(patch)
    ax.axis("off")

    if legend:
        if ax.legend_ is not None:
            ax.legend_.remove()
        legend = ax.legend(handles=legend_handles,
                           frameon=False, ncol=2, loc=2,
                           bbox_to_anchor=(0.0, -.05))
        ltext = legend.get_texts()
        plt.setp(ltext, fontproperties=textprops)
        ax.figure.subplots_adjust(hspace=.0)

    return ax, legend_handles


def plot_assembly_timeline(quote, deadline=None, ax=None,
                           rectangle_color="#bbbbff", scale=1.0,
                           backend="matplotlib"):
    """Make a Gantt-like chart of the assemblies."""
    if backend == "matplotlib":
        return _matplotlib_plot_assembly_timeline(
            quote, deadline=deadline, ax=ax, rectangle_color=rectangle_color,
            scale=scale
        )


def _matplotlib_plot_assembly_timeline(quote, deadline=None, ax=None,
                                       rectangle_color="#bbbbff", scale=1.0):
    """Plot the assembly timeline with Matplotlib."""
    if deadline is None:
        deadline = quote.lead_time
    if not hasattr(quote, 'tree'):
        quote.compute_full_assembly_tree()
    quote.propagate_deadline(deadline)
    assemblies_list = quote.tree_as_list()
    assemblies_list = sorted(assemblies_list, key=lambda a: a.id)

    textprops = fm.FontProperties(
        fname=SETTINGS["OpenSans-ttf-path"],
        size=12 * scale, family="sans-serif"
    )
    fontawesome = fm.FontProperties(
        fname=SETTINGS["fontawesome-ttf-path"],
        size=12 * scale, family="sans-serif"
    )
    if ax is None:
        fig, ax = plt.subplots(1, figsize=(16 * scale,
                                           len(assemblies_list) * 0.3 * scale))
    positions = {}

    if textprops is None:
        textprops = fm.FontProperties(
            fname=SETTINGS["OpenSans-ttf-path"],
            size=12, family="sans-serif"
        )

    for i, assembly in enumerate(assemblies_list):
        source = assembly.source
        color = "k"
        if source.__class__.__name__ == "CommercialDnaOffer":
            color = "#888888"
        x2 = assembly.deadline
        x1 = assembly.deadline - assembly.step_duration
        positions[assembly.id] = ([x1, i], [x2, i])
        patch = mpatches.Rectangle(
            (assembly.deadline - assembly.step_duration, i - .2),
            assembly.step_duration, .4, color=rectangle_color,
            ec="none"
        )

        ax.plot([0, assembly.deadline], [i, i], c="k", alpha=0.3,
                lw=.5 * scale, zorder=-1000, ls="-")
        ax.add_patch(patch)

        ax.text(-1.5, i, assembly.id, horizontalalignment="right",
                verticalalignment="center", fontproperties=textprops,
                color=color)
        ax.text(-.75, i, source.report_fa_symbol, horizontalalignment="center",
                verticalalignment="center", fontproperties=fontawesome,
                color=color)

    for assembly in assemblies_list:
        xa, ya = positions[assembly.id][0]
        if assembly.assembly_plan is not None:
            for segment, child in assembly.assembly_plan.items():
                xc, yc = positions[child.id][1]
                ax.plot([xc, xa], [yc, ya], lw=1 * scale, color=color)

    ax.set_ylim((-1, len(assemblies_list) + 1))
    ax.set_xlim(-2, deadline)
    ax.set_frame_on(False)
    ax.yaxis.set_visible(False)
    ticks = list(np.arange(0, deadline, deadline / 6).astype(int))

    ax.xaxis.set_ticks(ticks)
    ax.xaxis.set_ticks_position("bottom")
    ax.xaxis.set_tick_params(which='major', length=5 * scale)
    ax.xaxis.set_ticklabels(ticks,
                            fontproperties=textprops)
    ax.set_xlabel("TIME", fontproperties=textprops)
    ax.set_frame_on(False)
    return ax


def plot_tree_graph(levels, edges, draw_node, elements_positions=None,
                    ax=None, width_factor=2.5, height_factor=2, scale=1.0,
                    edge_left_space=0.015, edge_right_space=0.015,
                    interlevel_shift=0, margin=None, **txt_kw):
    """General function for plotting tree graphs.

    Parameters
    ----------

    levels
      A list of lists of nodes grouped by "level", i.e distance to the in the
      graph to the level 0. levels will be displayed on a same column.

    edges
      List of nodes pairs (source node, target node).

    draw_node
      A function f(x , y , node, ax, **kw) which draws something related to the
      node at the position x,y on the given Matplotlib ax.

    ax
      The matplotlib ax to use. If none is provided, a new ax is generated.

    Examples:
    ---------

    >>> def draw_node(x,y, node, ax):
        ax.text(x,y, node)
    >>> plot_tree_graph(levels=[["A","B","C"], ["D,E"], ["F"]],
                        edges=[("A","D"),("B","D"),("C","E")
                               ("D","F"),("E","F")],
                        draw_node = draw_node,)



    """
    levels_dict = {
        element: level
        for level, elements in enumerate(levels)
        for element in elements
    }
    if elements_positions is None:
        elements_positions = {}
        for lvl, elements in enumerate(levels):
            yy = np.linspace(0, 1, len(elements) + 2)[1:-1]
            yy += interlevel_shift * (1-2*(lvl % 2))
            x = 1.0 * (1 + lvl) / (len(levels) + 1)
            for y, element in zip(yy, elements):
                elements_positions[element] = (x, y)

    if ax is None:
        width = width_factor * len(levels) * scale
        height = height_factor * max([len(lvl) for lvl in levels]) * scale
        fig, ax = plt.subplots(1, figsize=(width, height))

    for element, (x, y) in elements_positions.items():
        draw_node(x, y, element, ax, **txt_kw)

    y_spans = [
        elements_positions[elements[1]][1] - elements_positions[elements[0]][1]
        for elements in levels
        if len(elements) > 1
    ]

    delta_y = 0.5*min(y_spans) if y_spans != [] else 0

    for el1, el2 in edges:
        x1, y1 = elements_positions[el1]
        x2, y2 = elements_positions[el2]
        x1 += edge_left_space * np.sqrt(scale)
        x2 += -edge_right_space * np.sqrt(scale)
        if ((levels_dict[el2] - levels_dict[el1]) > 1) and (y1 == y2):
            patch = mpatches.PathPatch(
                Path([(x1, y1), (0.5 * x2 + 0.5 * x1, y1-delta_y),
                      (0.5 * x2 + 0.5 * x1, y2-delta_y), (x2, y2)],
                     [Path.MOVETO, Path.CURVE4, Path.CURVE4, Path.CURVE4]),
                facecolor='none', lw=1 * scale,
                zorder=-1000
            )

        else:
            patch = mpatches.PathPatch(
                Path([(x1, y1), (0.9 * x2 + 0.1 * x1, y1),
                      (0.1 * x2 + 0.9 * x1, y2), (x2, y2)],
                     [Path.MOVETO, Path.CURVE4, Path.CURVE4, Path.CURVE4]),
                facecolor='none', lw=1 * scale,
                zorder=-1000
            )

        ax.add_patch(patch)

    ax.axis("off")
    if margin is not None:
        xx, yy = [np.array(e) for e in zip(*elements_positions.values())]
        xmin, xmax = xx.min(), xx.max()
        dx = margin * (xmax - xmin)
        ymin, ymax = yy.min(), yy.max()
        dy = margin * (ymax - ymin)
        d = max(dx, dy)
        ax.set_xlim(xmin - d, xmax + d)
        ax.set_ylim(ymin - d, ymax + d)

    return elements_positions, ax


def plot_supply_graph(quote, ax=None, textprops=None, margin=None,
                      scale=1.0, interlevel_shift=0):
    """Plot the supply graph of all sources related to the provided source.

    Examples:
    ---------

    >>> quote.sources = chunks_assembly_station.compute_dict_supply_graph()

    Parameters
    ----------

    quote.sources
      A dictionary {"source_1_name": {"depth": 1, "infos": {...}},
                    "source_2_name": {"depth": 2, "infos": {...}},
                    etc.}
    ax
      A matplotlib ax

    textprops
      Properties of the text. If none, a nice opensans font is used.


    """
    edges = []
    levels = defaultdict(lambda *a: [])
    for source_name, infos in quote.sources.items():
        levels[infos._depth].append(source_name)
        for provider_name in infos.providers:
            edges.append((provider_name, source_name))
    levels = [levels[i] for i in range(max(levels) + 1)][::-1]

    fontawesome = fm.FontProperties(
        fname=SETTINGS["fontawesome-ttf-path"],
        size=18 * scale, family="sans-serif"
    )
    if textprops is None:
        textprops = fm.FontProperties(
            fname=SETTINGS["OpenSans-ttf-path"],
            size=12 * scale, family="sans-serif"
        )

    def draw_node(x, y, source_name, ax):
        source_infos = quote.sources[source_name]
        symbol = source_infos._report_fa_symbol
        ax.text(x, y, symbol, horizontalalignment="center",
                verticalalignment="center", fontproperties=fontawesome)

        text = "\n".join(textwrap.wrap(source_infos.name, 13))

        ax.text(x, y + 0.035 * scale**.3, text,
                horizontalalignment="center",
                verticalalignment="bottom", fontproperties=textprops,
                )
    return plot_tree_graph(levels, edges, draw_node, ax=ax, scale=scale,
                           interlevel_shift=interlevel_shift, margin=margin)


def plot_assembly_graph(quote, ax=None, margin=None, textprops=None,
                        scale=1.0):
    """Plot the complete assembly graph"""

    nodes_dict = {}
    levels = defaultdict(lambda *a: [])
    edges = []
    if not hasattr(quote, 'tree'):
        quote.compute_full_assembly_tree()
    tree = deepcopy(quote.tree)

    def rec(node, depth=0):
        if node.get("_visited", False):
            return
        nodes_dict[node.id] = node
        node["_visited"] = True
        assembly_plan = node.pop("assembly_plan")
        levels[depth].append(node.id)
        for other in assembly_plan:
            edges.append([other.id, node.id])
            rec(other, depth + 1)
    rec(tree)
    levels = [levels[i] for i in range(max(levels) + 1)][::-1]

    fontawesome = fm.FontProperties(
        fname=SETTINGS["fontawesome-ttf-path"],
        size=13 * scale, family="sans-serif"
    )
    if textprops is None:
        textprops = fm.FontProperties(
            fname=SETTINGS["OpenSans-ttf-path"],
            size=12 * scale, family="sans-serif"
        )

    def draw_node(x, y, node_id, ax):
        node = nodes_dict[node_id]
        icon = quote.sources[node.source]._report_fa_symbol
        ax.text(x, y, node_id, horizontalalignment="left",
                verticalalignment="center",
                fontproperties=textprops)
        ax.text(x - 0.01 * np.sqrt(scale), y, icon,
                horizontalalignment="right",
                verticalalignment="center",
                fontproperties=fontawesome)

    all_elements = sorted(sum(levels, []))
    ypos = {
        el: 1.0 * (i + 1) / (len(all_elements) + 2)
        for i, el in enumerate(all_elements)
    }
    for el in all_elements:
        children = [
            e2
            for (e2, e1) in edges
            if e1 == el
        ]
        if children != []:
            ypos[el] = 1.0 * sum(ypos[e] for e in children) / len(children)

    xpos = {
        el: 1.0 * (1 + x) / (len(levels) + 1)
        for x, elements in enumerate(levels)
        for el in elements
    }

    elements_positions = {el: (xpos[el], ypos[el]) for el in all_elements}
    return plot_tree_graph(levels, edges, draw_node,
                           elements_positions=elements_positions, ax=ax,
                           edge_left_space=0.06,
                           edge_right_space=0.03,
                           margin=margin,
                           height_factor=0.40, width_factor=5.5, scale=scale)


def matplotlib_figure_to_file_string(fig, format="svg", **kwargs):
    """Return a string of the figure in the requested format."""

    if (format == "pdf") and USE_BYTES:
        output = BytesIO()
    else:
        output = StringIO()

    fig.savefig(output, format=format, **kwargs)
    return output.getvalue()


def matplotlib_figure_to_svg_base64_data(fig, **kwargs):
    """Return a string of the form 'data:image/svg+xml;base64,XXX' where XXX
    is the base64-encoded svg version of the figure."""
    svg_txt = matplotlib_figure_to_file_string(fig, format="svg", **kwargs)
    svg_txt = "\n".join(svg_txt.split("\n")[4:])
    svg_txt = "".join(svg_txt.split("\n"))
    try:
        return "data:image/svg+xml;base64," + b64encode(svg_txt)
    except:
        content = b64encode(svg_txt.encode("ascii"))
        result = (b"data:image/svg+xml;base64," + content).decode("utf-8")
        return str(result)


def give_quotes_html_locations(quotes, len_sequence, ax=None):
    """Quickly adds an html-location field containing an HTML image of the
    location of each sub-quote in the final sequence.

    The image is a line representing the full sequence and a rectangle
    indicating the position of the subquote in the the full sequence.

    The advantage of processing all quotes at once is that the
    figure is drawn once, then only the position of the rectangle and the
    title are modified.
    """
    if ax is None:
        fig, ax = plt.subplots(1, figsize=(1.5, 0.2), facecolor=None)
    ax.set_frame_on(False)
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    ax.plot([0, len_sequence], [0, 0], c="k")
    rect = mpatches.Rectangle((0, -0.5), 0, 1, color="#ddddff",
                              ec="k", alpha=1, zorder=1000,
                              transform=ax.transData)
    ax.add_patch(rect)
    ax.set_ylim((-1, 1))
    ax.set_xlim(0, len_sequence)
    text = ax.text(0.5, 0.65, "none", horizontalalignment="center", fontsize=4)
    for q in quotes:
        location = q.final_location
        if location is None:
            q.html_location = ''
        else:
            rect.set_bounds(location[0], -.5, location[1] - location[0], 1.0)
            text.set_text("%d-%d" % (location[0], location[1]))
            svg_data = matplotlib_figure_to_svg_base64_data(ax.figure,
                                                            transparent=True)
            q.html_location = "<img src='%s' />" % svg_data
    plt.close(ax.figure)


def hls_to_hex(hue, luminance, saturation):
    """Return (R,G,B) equivalent of a hue/staturation/value color."""
    return cl.rgb2hex(colorsys.hls_to_rgb(hue, luminance, saturation))

def rgb_to_hex(red, green, blue):
    """Return color as #rrggbb for the given color values."""
    return '#%02x%02x%02x' % (int(red), int(green), int(blue))

def autocolor_quote_sources(quote, hues=(0.635, 0.047, 0.117),
                            saturations=(0.9, 0.7, 0.5, 0.3),
                            min_lum=0.2, max_lum=0.8):
    """Auto-add a `_report_color` field to the sources in in quote.sources.

    Sources at the same depth share the same luminance
    """

    colors = itertools.cycle([
        rgb_to_hex(*[
            255*e**0.4
            for e in cm.Paired(0.13 * i % 1.0)
        ][:3])
        for i in range(30)
    ])
    for name, source in sorted(quote.sources.items()):
        color = next(colors)
        source._report_color = color

def ax_to_pdf(ax):
    return lambda fh: ax.figure.savefig(fh, format="pdf", bbox_inches="tight")

def plot_decomposition_graph(graph, nodes_color="#6886b7", weight='weight',
                             colormap='jet', ax=None, edge_width=1,
                             edges_step=1, edge_alpha=0.2, figsize=(8, 8)):
    xx = np.array(sorted(graph.nodes))
    L = xx.max()
    edges = [
        (start, end, data[weight] / (end - start))
        for start, end, data in graph.edges(data=True)
    ]
    edges = edges[::edges_step]
    edges = sorted(edges, key=lambda e: e[2])
    max_segment_length = max([end - start for (start, end, _) in edges])
    weights = np.array([w for (_, _, w) in edges])
    normalized_weights = (255 * weights / weights.max()).astype('uint8')
    colormap = cm.__dict__[colormap]
    colors = colormap(normalized_weights, alpha=edge_alpha)
    if ax is None:
        fig, ax = plt.subplots(1, figsize=figsize)
    ax.axis("off")
#     print (list(colors))
    for (start, end, w), color in zip(edges, colors):
        xc = 0.5 * (start + end)
        half = 0.5 * abs(end - start)
        ax.add_patch(mpatches.Arc((xc, 0), 2 * half, 2 * half,
                                  theta1=0 ,theta2=180, facecolor='none',
                                  ls='-', edgecolor=color, linewidth=1))
    ax.plot(xx, [0 for x in xx], marker="o", c=nodes_color)
    ax.set_aspect("equal")
    ax.set_xlim(-1, L + 1)
    ax.set_ylim(-1, max_segment_length / 2 + 1)

def plot_cost_profiles(cost_profiles):
    """Plot cost profiles generated in OptimizeManufacturability"""
    fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True)
    L = len(cost_profiles['count'])
    diffs = cost_profiles['diffs']
    r = (L - len(diffs)) / 2
    xx = range(L)
    ax1.plot(r + np.arange(len(diffs)), diffs, c='orange')
    ymax = diffs.max()
    if ymax == 0:
        ymax = 1
    ax1.set_ylim(bottom=-0.005, top = 1.5 * ymax)
    ax1b = ax1.twinx()
    ax1b.plot(xx, cost_profiles['count'], c='k', alpha=0.3)
    ax1b.set_ylim(bottom=0, top = 1.5 * cost_profiles['count'].max())
    ax1.set_xlim(0, L)
    
    ax2.plot(xx, cost_profiles['mean'], c='b', lw=2)
    ax2.fill_between(
        xx,
        cost_profiles['min'],
        cost_profiles['max'],
        facecolor='b',
        alpha=0.2
    )
    ax2.set_ylim(bottom=0, top = 1.5 * cost_profiles['max'].max())
    return ax1, ax2